/* $Id: camera.h,v 1.18 2007/10/11 12:42:47 cvsroot Exp $ */

/* DGAP - General Analytical Positioning
 * Copyright (C) 2005 Dirk Stallmann
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#ifndef DEP_CAMERA_INCLUDED
#define DEP_CAMERA_INCLUDED

#include "depdef.h"
#include "drvsub.h"

using namespace std;
using namespace GenLib2;

namespace DEP
{
class ImagePoint;
class APSet;
class LinObsEqn;
class LinConditionEqn;
class Params;

/** Camera class.
 *  Implementation of a ''pure'' frame camera.
 *  This class contains the three parameters of the inner
 *  orientation (IO): principal distance and principal point coordinates.
 *  The concept of IO follows the Manual of Photogrammetry
 *  (Slama et al., 1980).
 *
 *  @author Dirk Stallmann
 */
class Camera
{
public:
	/// types of camera coordinate systems
	enum CSystem
	{
		IMAGE,		/* image coordinate system with coordinates in meter */
		PIXEL		/* lefthanded pixel coordinate system.
						Origin is at top left corner, x-axis to the right,
						y-axis downwards. The center of the top left pixel
						has the coordinates (0.0, 0.0). Coordinates are in pixel. */
	};

	/// constructor
	Camera();
	
	/// destructor
	virtual ~Camera() {}
	
	/**@name set/get member data */
	//@{
		/// set camera ID
		void set_id(const char* id) { _id = id; }
	
		/// get camera ID
		const char* get_id() const { return _id.c_str(); }	
	
		/// get camera type
		virtual string type_str() const;
	
		/// get keyword name
		virtual string key_str() const;
		
		/// get type of coordinate system type
		virtual string csys_str() const;
		
		/// get type of coordinate system
		virtual CSystem csystem() const { return IMAGE; }
		
		/// get principal distance
		double ck() const { return _princ_dist; }

		/// get principal point coords.
		const DPoint2& get_pp() const { return _princ_pnt; }
	
		/// get image size in x (width)
		double get_size_x() const { return _size_x; }
	
		/// get image size in y (height)
		double get_size_y() const { return _size_y; }
	
		/// set inuse flag
		void inuse(bool flg) { _inuse = flg; }
	
		/// test inuse flag
		bool is_unused() const { return _inuse == false; }
	//@}
	
	/**@name read/write/print camera parameters
	 */
	//@{
		/// read camera parameters
		virtual bool read(istream& is);
	
		/// write camera parameters
		virtual void write(ostream& os) const;
	
		/// print camera parameters
		virtual void print(ostream& os) const;
	//@}
	
	/**@name sensor / image coordinates transformation
	 *  To convert between sensor and image coordinates.
	 *  In this special case sensor and image coordinates
	 *  are identical!
	 *  A derived class which implements e.g. a CCD camera will
	 *  override this functions to provide the pixel / image coordinate
	 *  transformation.
	 */
	//@{
		/// sensor-to-image coordinates transformation
		virtual void sensor_to_image(const DPoint2& sen_c,
			DPoint2& img_c) const { img_c = sen_c; }

		/// image-to-sensor-coords. transformation
		virtual void image_to_sensor(const DPoint2& img_c,
			DPoint2& sen_c) const { sen_c = img_c; }
	//@}

	/** transformation of standard deviations
	 *  from sensor to image coordinates
	 */
	virtual void sdev_sensor_to_image(const DPoint2& sen_s,
		DPoint2& img_s) const { img_s = sen_s; }
	
	/**@name image / photo coordinates transformation
	 *  To convert between image and photo coordinates.
	 *  Photo coordinates are image coordinates reduced to the principal
	 *  point. Corrections of APs are included!
	 */
	//@{
	 	/// image-to-photo coordinates transformation
		void image_to_photo(const DPoint2& img_c, DPoint3& pho_c,
			bool corr_flg=true) const;

		/// photo-to-image-coords. transformation
		void photo_to_image(const DPoint3& pho_c, DPoint2& img_c,
			bool corr_flg=true) const;
	//@}

	/**@name read/write/print sensor coordinates.
	 */
	//@{
		/// read sensor coords.
		virtual bool read_sensor_coords(istream& is,
			vector<ImagePoint*>& impts, bool sdev_flg, const DPoint2& sdev) const;
	
		/// write sensor coords.
		virtual void write_sensor_coords(ostream& os,
			const vector<ImagePoint*>& impts) const;
	
		/// print sensor coords.
		virtual void print_sensor_coords(ostream& os,
			const vector<ImagePoint*>& impts, bool ic_out) const;
	//@}
	
	/**@name AP related functions
	 */
	//@{
		/// attach AP set
		void attach(APSet* aps) { _ap_sets.push_back(aps); }

		/// check link to AP set
		bool has_ap_set() const { return !_ap_sets.empty(); }

		/// get AP set correction
		DPoint2 ap_set_corr(const DPoint3& pho_c) const;

		/// add subset of derivatives to linear observation equations
		void add(LinObsEqn& eqn) const;

		/// add subset of derivatives to linear condition equations
		void add_a(LinConditionEqn& eqn) const;

		/// get number of AP sets
		size_t num_ap_sets() const { return _ap_sets.size(); }
		
		// get number of APs
		//size_t num_aps() const;
		
		/// get partial derivatives of IO parameters
		void update_derivs(const DPoint3& pho_c);

		/// get partial derivatives of IO parameters
		void get_derivs(const DPoint3& pho_c, vector<DerivSub>& derivs) const;

		/// print influence of APs
		void print_aps_influence(ostream& os, const Params& pr) const;

		/// update IO parameter using optional AP set
		bool update_io();
	//@}

protected:
	/// camera ID
	string		_id;

	/// comment
	string		_comment;
	
	/// image size in x (width)
	double		_size_x;

	/// image size in y (height)
	double		_size_y;

	/// inuse flag
	bool		_inuse;
	
	/// AP set list (maybe empty)
	vector<APSet*>	_ap_sets;

	/// principal distance (= camera constant)
	double		_princ_dist;

	/// principal point coordinates
	DPoint2		_princ_pnt;
};


// non-member functions ...

} // namespace DEP

#endif
