/* $Id: ccdcam.cpp,v 1.12 2007/10/11 12:44:54 cvsroot Exp $ */

/* DGAP - General Analytical Positioning
 * Copyright (C) 2005 Dirk Stallmann
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include "ccdcam.h"
#include "image.h"
#include "imgpnt.h"

namespace DEP
{

// constructor

CCDCamera::CCDCamera() : Camera(),
	_nc(0),
	_nr(0),
	_sc(0.),
	_sr(0.),
	_c0(0.),
	_r0(0.)
{
	//_csystem = PIXEL;
}


// get camera type

string CCDCamera::type_str() const
{
	ostringstream oss;
	oss << "CCD-frame";
	return oss.str();
}


// get keyword name

string CCDCamera::key_str() const
{
	ostringstream oss;
	oss << "ccd-camera";
	return oss.str();
}


// get coordinate system type

string CCDCamera::csys_str() const
{
	ostringstream oss;
	oss << "pixel";
	return oss.str();
}


// get orientation type

string CCDCamera::orientation_str() const
{
	ostringstream oss;
	oss << "top-left";
	return oss.str();
}


// read camera parameters

bool CCDCamera::read(istream& is)
{
	// identifier
	is >> _id;
	if (!is) return false;

	// comment
	is >> _comment;
	if (!is) return false;

	// principal distance and principal point in mm
	is >> _princ_dist >> _princ_pnt;
	if (!is) return false;

	// convert from mm to m
	_princ_pnt *= 1e-3;
	_princ_dist *= 1e-3;

	// sensor size and pixel spacing
	is >> _nc >> _nr >> _sc >> _sr;
	if (!is) return false;

	// convert from um/pixel to m/pixel
	_sc *= 1e-6;
	_sr *= 1e-6;

	// calc. ICS location
	_c0 = (_nc - 1) / 2.;
	_r0 = (_nr - 1) / 2.;

	// calc. image format
	_size_x = _nc * _sc;
	_size_y = _nr * _sr;

	return true;
}


// write camera parameters

void CCDCamera::write(ostream& os) const
{
	Form fix3(3); fix3.fixed().width(6);
	Form fix6(6); fix6.fixed().width(12);

	os << "camera-ccd   " << _id << "   " << _comment << endl;
	os << fix6(_princ_dist * 1e3) << " "
		<< fix6(_princ_pnt.x * 1e3) << " " << fix6(_princ_pnt.y * 1e3) << endl;
	os << _nc << " " << _nr << endl;
	os << fix3(_sc * 1e6) << " " << fix3(_sr * 1e6);
	os << endl;
}


// print camera parameters

void CCDCamera::print(ostream& os) const
{
	Form int10; int10.width(10);
	Form fix3(3); fix3.fixed().width(10);

	Camera::print(os);

	os << "Number of pixel    :  nc = " << int10(_nc) << " pix     "
		<< "nr = " << int10(_nr) << " pix\n";
	os << "ICS location       :  c0 = " << fix3(_c0) << " pix     "
		<< "r0 = " << fix3(_r0) << " pix\n";
	os << "Pixel spacing      :  sc = " << fix3(_sc * 1e6) << " um/pix  "
		<< "sr = " << fix3(_sr * 1e6) << " um/pix\n";
	os << endl;
}


// sensor-to-image-coords. transformation

void CCDCamera::sensor_to_image(const DPoint2& sen_c, DPoint2& img_c) const
{
	//cout << "CCDCamera::sensor_to_image" << endl;
	//img_c.x = (pix_c.x - _cp) * _sc;
	//img_c.y = (_rp - pix_c.y) * _sr;
	img_c.x =   (sen_c.x - _c0) * _sc;
	img_c.y = - (sen_c.y - _r0) * _sr;
}


// image-to-sensor-coords. transformation

void CCDCamera::image_to_sensor(const DPoint2& img_c, DPoint2& sen_c) const
{
	//cout << "CCDCamera::image_to_sensor" << endl;
	sen_c.x = _c0 + img_c.x / _sc;
	sen_c.y = _r0 - img_c.y / _sr;
}


// transformation of standard deviations from sensor to image coordinates

void CCDCamera::sdev_sensor_to_image(const DPoint2& sen_s,
	DPoint2& img_s) const
{
	img_s.x = sen_s.x * _sc;
	img_s.y = sen_s.y * _sr;
}


// read sensor coords.

bool CCDCamera::read_sensor_coords(istream& is,
	vector<ImagePoint*>& impts, bool sdev_flg, const DPoint2& sdev) const
{
	// read image offset
	double off_x = 0., off_y = 0.;
	is >> off_x >> off_y;
	if (!is) return false;

	string id;

	while (is >> id)
		{
		if (id.compare("stop-dep") == 0 || id.compare("-1") == 0)
			break; // stop reading

		ImagePoint *ip = new ImagePoint;
		assert(ip != 0);
		ip->set_id(id);

		// read image coords.
		is >> ip->crds;
		if (!is) return false; // read error
	
		// std. devs.
		if (sdev_flg)
			ip->sdev = sdev;
		else
			{
			is >> ip->sdev;
			if (!is) return false;
			}
			
		// translate coords.
		ip->crds.x += off_x;
		ip->crds.y += off_y;

		// append or replace point
		add_image_point(impts, ip);
		}

	return true;
}


// write sensor coords.

void CCDCamera::write_sensor_coords(ostream& os,
const vector<ImagePoint*>& impts) const
{
	assert(false);
	/*
	// scale/unit should be selectable by LS-params./options!
	const double sc = 1e-6;

	Form str8; str8.width(8);
	Form fix8(4); fix8.fixed().width(8);
	Form fix10(4); fix10.fixed().width(10);

	os << sc << endl;

	vector<ImagePoint*>::const_iterator ip;
	for (ip = impts.begin(); ip != impts.end(); ++ip)
		{
		if (*ip == 0) continue;
		os << str8((*ip)->get_id())
			<< " " << fix10((*ip)->crds.x * sc)
			<< " " << fix10((*ip)->crds.y * sc)
			<< " " << fix8((*ip)->sdev.x * sc)
				<< " " << fix8((*ip)->sdev.y * sc) << endl;
		}
	os << endl;
	*/
}


// print sensor coords.

void CCDCamera::print_sensor_coords(ostream& os,
const vector<ImagePoint*>& impts, bool ic_out) const
{
	Form str8; str8.width(8);
	Form fix8(3); fix8.fixed().width(8);
	Form fix10(3); fix10.fixed().width(10);
	
	os << "Number of points        :  "
		<< DEP::count_image_points(impts) << endl;
	os << "Coord. system type      :  " << csys_str() << endl;
	os << "Orientation             :  " << orientation_str() << endl;
	os << endl;
	
	if (ic_out == false) return; // escape
	
	os << "      ID     x/pix      y/pix    sx/pix   sy/pix\n\n";
	vector<ImagePoint*>::const_iterator ip;
	for (ip = impts.begin(); ip != impts.end(); ++ip)
		{
		if (*ip == 0) continue;
		os << str8((*ip)->get_id())
			<< " " << fix10((*ip)->crds.x)
			<< " " << fix10((*ip)->crds.y)
			<< " " << fix8((*ip)->sdev.x)
			<< " " << fix8((*ip)->sdev.y) << endl;
		}
	os << endl;
}


// calculate GSD

void CCDCamera::GSD(double hg, double& gsd_c, double& gsd_r)
{
	double m = hg / _princ_dist;
	gsd_c = m * _sc;
	gsd_r = m * _sr;
}

// non-member functions ...

} // namespace DEP
