/* $Id: dlt.h,v 1.11 2007/12/10 06:53:49 cvsroot Exp $ */

/* DGAP - General Analytical Positioning
 * Copyright (C) 2005 Dirk Stallmann
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#ifndef DEP_DLT_INCLUDED
#define DEP_DLT_INCLUDED

#include "image.h"
#include "imggeo.h"

using namespace std;
using namespace GenLib2;

namespace DEP
{

/** DLT parameter class.
 *  This class holds the Direct Linear Transformation (DLT) parameters
 *  L1, ... L11.
 *
 *  @author Dirk Stallmann
 */
class DLTParameters : public TransfParams
{
public:
	/// constructor
	DLTParameters() : TransfParams(11) {}
	
	/**@name overridden base class AdjObject members */
	//@{
		/// get type name
		string type_str() const;

		/// get key name
		string key_str() const;

		/// get parameter name
		string uname(size_t i) const;
	//@}

	/// calculate denominator
	double denom(const DPoint3& obj_c) const
	{
		return (_a[8] * obj_c.x + _a[9] * obj_c.y + _a[10] * obj_c.z + 1.);
	}
	
	/// object-to-image coords. transformation
	bool object_to_image(const DPoint3& obj_c, DPoint2& img_c) const;
	
	/** IO and EO parameters calculated from DLT parameters
	 *  see Non-Topographic Photogrammetry page 47 (4.43) and
	 *  Bopp&Kraus, PEARS Vol. 44, No. 9, 1978, pp. 1191-1196.
	 */
	bool calc_orient_from_dlt(DPoint2& pp, DPoint2& ck, DPoint3& pcc,
		DRotMatrix3& R) const;
	
	/// Sarrus formular
	double sarrus(double a[3][3]) const;
};


/** Direct Linear Transformation class.
 *  Solution of the analytical orientation problem using
 *  the method of Direct Linear Transformation (DLT).
 *  Iterative solution approach of the DLT.
 *
 *  @author Dirk Stallmann
 */
class DLTImage : public Image
{
public:
	/// constructor
	DLTImage();
	
	/// get image type
	virtual string type_str() const;
	
	/**@name object-to-image coords. transformation
	 *
	 *  @param obj_c (IN) object coordinates
	 *  @param img_c (OUT) image coordinates
	 *  @param r (IN) row/scanline coordinate (UNSED)
	 *  @return true = no problem, false = error encountered
	 */
	virtual bool object_to_image(const DPoint3& obj_c,
		DPoint2& img_c, double r=0) const;
	
	/// resection using the direct solution approach
	virtual bool resection(ostream& os, const vector<ObjectPoint*>& obpts, bool dbg);
	
	/** observation equation for intersection
	 *  See Schmid, BuL 1958, Nr 4, Seite 112-113.
	 *
	 *  @param id (IN) object points ID
	 *  @param a (OUT) coeffs. for design matrix A
	 *  @param wt (OUT) misclosures
	 *  @param found (OUT) true = point found, false = point not found
	 *  @return true = no problem, false = error encountered
	 */
	virtual bool observ_eqn_intersec(const string& id, DerivSub& a, DVector& wt,
		bool& found) const;
	
	/// add constribution to normal equations for all image point observations
	virtual bool normal_eqns(const vector<ObjectPoint*>& obpts,
		Matrix<double>& atpa, Vector<double>& atpl, double& ltpl,
		int& num_obs) const;
		
	/// IO and EO parameters calculated from DLT parameters
	void print_orient_from_dlt(ostream& os) const;
};


// non-member functions ...

/** Project object point onto image plane using DLT.
 *  Basic DLT equation to get image from object coordinates.
 */
void dlt_projection(const double* l, const DPoint3& obj_c, DPoint2& img_c);

} // namespace DEP

#endif
