/**********************************************************************/
//
// FILE:    frameext.cpp
//          Extened frame image class functions.
// CREATED: 23 Feb 05 (ds)
// CHANGES: see include file
// REFS:    dito 
//
// (c) 2005 Dirk Stallmann (ds)/ IfP Uni Stuttgart
/**********************************************************************/

#include "depvec.h"
#include "frameext.h"
//#include "lspars.h"
#include "drvsub.h"
#include "lconeqn.h"
//#include "lobseqn.h"
#include "imgpnt.h"
#include "objpnt.h"
#include "camera.h"
#include "extori.h"
//#include "apset.h"
#include "poscor.h"
#include "attcor.h"
#include "imumis.h"
//#include "iopntcon.h"

namespace DEP
{

// constructor

ExtFrameImage::ExtFrameImage() : Image(),
	_imu_mis(0)
{
	// empty
}


// get image type

string ExtFrameImage::type_str() const
{
	ostringstream oss;
	oss << "frame extended";
	return oss.str();
}


// calculate exterior orientation

bool ExtFrameImage::calc_ext_orientation(DPoint3& pcc, DRotMatrix3& N) const
{
	assert(_station != 0);
	
	const ExtOrientation* eo = dynamic_cast<const ExtOrientation*>(_station);
	assert(eo != 0);
	
	// perspective centre coords.
	eo->get_per_centre(pcc);
	
	// exposure time
	double t = eo->get_time();
	
	// position crrection (optional)
	if (_pos_cor != 0)
		pcc += _pos_cor->calc_corr(t);
	
	// rotation parameters
	size_t n = eo->num_rot_params(); // test missing!
	vector<double> a(n);
	eo->get_rot_params(a);
	
	// attitude correction (optional)
	if (_att_cor != 0)
		{
		DPoint3 da = _att_cor->calc_corr(t);
		a[0] += da.x;
		a[1] += da.y;
		a[2] += da.z;
		}
		
	// rotation matrix
	DRotMatrix3 R;
	eo->calc_rot_matrix(R, a);
	
	// IMU misalignment (optional)
	DRotMatrix3 M;
	if (_imu_mis != 0)
		_imu_mis->get_rot_matrix(M);
	
	// total rotation matrix
	// attitude correction and misalignment included
	N = R * M;
	//cout << "ExtFrameImage" << endl;
	
	return true;
}


// // image-to-object coords. transformation
// 
// bool ExtFrameImage::image_to_object(const DPoint2& img_c, DPoint3& obj_c) const
// {
// 	cout << "ERROR: image_to_object() not implemented yet!" << endl;
// 	assert(false);
// 
// 	return true;
// }


// add contribution to normal equations for all image point observations
// using the Gauss-Helmert model

bool ExtFrameImage::normal_eqns(const vector<ObjectPoint*>& obpts,
Matrix<double>& atpa, Vector<double>& atpl, double& ltpl, int& num_obs) const
{
	assert(_station != 0);	// necessary
	assert(_camera != 0);	// necessary
	
	const ExtOrientation* eo = dynamic_cast<const ExtOrientation*>(_station);
	assert(eo != 0);

	// Setup sub-matrices of derivatives
	// Always 2 observation equations for each image point.
	// The number of columns might be variable.
	DerivSub	deriv_oc(2,3);		// object points, size always 2x3
	DerivSub	deriv_pcc(2,3);		// per. centre coords., size always 2x3
	DerivSub	deriv_rot(2,eo->num_rot_params());		// rotation params., variable size
	//vector<DerivSub> deriv_ap(_ap_sets.size()); // APs, variable size
	vector<DerivSub> deriv_ap(_camera->num_ap_sets());	// APs, variable size
	DerivSub	deriv_pc;			// position corr., variable size
	DerivSub	deriv_ac;			// attitude corr., variable size
	DerivSub	deriv_mis;			// IMU misalignment, variable size
	DerivSub	deriv_obs(2,2);		// observations, size always 2x2

	// derivatives of observations
	deriv_obs.identity();
	
	// set index for EO (valid for the whole image)
	deriv_pcc.index(eo->get_index());
	deriv_rot.index(eo->get_index() + 3);

	// always 2 condition equations for a single imaging ray.
	LinConditionEqn con_eqn;

	// design matrix A
	con_eqn.add_a(&deriv_oc);	// always
	con_eqn.add_a(&deriv_pcc);	// always
	con_eqn.add_a(&deriv_rot);	// always
	con_eqn.add_a(deriv_ap);	// optional
	if (_pos_cor != 0) con_eqn.add_a(&deriv_pc);	//optional
	if (_att_cor != 0) con_eqn.add_a(&deriv_ac);	//optional
	if (_imu_mis != 0) con_eqn.add_a(&deriv_mis);	//optional
	
	// perspective centre coords.
	DPoint3 pcc;
	eo->get_per_centre(pcc);
	
	// exposure time
	double t = eo->get_time();
	
	// position correction (optional)
	if (_pos_cor != 0)
		pcc += _pos_cor->calc_corr(t);
	//cout << pcc << endl;
		
	// rotation parameter
	size_t n = eo->num_rot_params();
	vector<double> a(n);
	eo->get_rot_params(a);
		
	// attitude correction (optional)
	if (_att_cor != 0)
		{
		DPoint3 da = _att_cor->calc_corr(t);
		a[0] += da.x;
		a[1] += da.y;
		a[2] += da.z;
		}
		
	DRotMatrix3 R; // instantenous rotation matrix
	DRotMatrix3 M; // IMU misalignment matrix
	DRotMatrix3 N; // rotation matrix: N = R M
	
	// rotation matrix
	eo->calc_rot_matrix(R, a);
	//cout << "R =\n" << R << endl;
	
	// IMU misalignment matrix (optional)
	if (_imu_mis != 0)
		_imu_mis->get_rot_matrix(M);
	//cout << "M =\n" << M << endl;
	
	// total rotation matrix: N = R M,
	// attitude correction and misalignment included
	N = R * M;
	//cout << "N =\n" << N << endl;
	
	// loop over all image points
	vector<ImagePoint*>::const_iterator ip;
	for (ip = _image_points.begin(); ip != _image_points.end(); ++ip)
		{
		if (*ip == 0) continue;
		
		// search for corresponding object point
		ObjectPoint* op = find(obpts, (*ip)->get_id());
		if (op == 0)
			{
			cout << "ATTENTION: Object point " << (*ip)->get_id()
				<< " not found.\n";
			continue;
			}

		//cout << _image_id << ":" << (*ip)->get_id() << ":\n"; cout.flush();

		// set current index of object point
		deriv_oc.index(op->get_index());

		// convert sensor to image std. dev.
		DPoint2 sdev;
		_camera->sdev_sensor_to_image((*ip)->sdev, sdev);
		//cout << (*ip)->get_id() << ": " << sdev << endl;

		// co-factors
		Matrix<double> Qll(2,2);
		Qll(0,0) = sdev.x / AdjObject::lspars.sigma0(); Qll(0,0) *= Qll(0,0);
		Qll(1,1) = sdev.y / AdjObject::lspars.sigma0(); Qll(1,1) *= Qll(1,1);
		con_eqn.set_q(Qll);
		//cout << "Qll: " << Qll << endl;

		// convert sensor into image coords.
		DPoint2 img_c;
		_camera->sensor_to_image((*ip)->crds, img_c);
		//cout << "img: " << img_c << endl;

		// convert image into photo coords., AP included
		DPoint3 pho_c;
		image_to_photo(img_c, pho_c);
		//cout << "pho: " << pho_c << endl;

		// partial derivatives of object point coords.
		// and orientation parameters
		// The derivatives of the projection centre coordinates and
		// the object coordinates are always the same only the derivatives
		// of the rotation parameters are different!

		// object coords. reduced to the perspective centre
		DPoint3 d;
		d.x = op->crds.x - pcc.x;
		d.y = op->crds.y - pcc.y;
		d.z = op->crds.z - pcc.z;

		double w = N(0,2) * d.x + N(1,2) * d.y + N(2,2) * d.z;
		if (fabs(w) == 0.)
			{
			cout << "ERROR: ExtFrameImage::normal_eqns: Zero denominator.\n";
			return false;
			}

		double u = N(0,0) * d.x + N(1,0) * d.y + N(2,0) * d.z;
		double v = N(0,1) * d.x + N(1,1) * d.y + N(2,1) * d.z;

		double g = pho_c.z / w;
		double h = g / w;
	
		// derivatives of the object coordinates (always)
		// X-derivatives
		deriv_oc(0,0) = h * (N(0,0) * w - N(0,2) * u);
		deriv_oc(1,0) = h * (N(0,1) * w - N(0,2) * v);

		// Y-derivatives
		deriv_oc(0,1) = h * (N(1,0) * w - N(1,2) * u);
		deriv_oc(1,1) = h * (N(1,1) * w - N(1,2) * v);

		// Z-derivatives
		deriv_oc(0,2) = h * (N(2,0) * w - N(2,2) * u);
		deriv_oc(1,2) = h * (N(2,1) * w - N(2,2) * v);
	
		// The derivatives of the projection centre coordinates are simply
		// the negative derivatives of the object coordinates !
		// Xo-derivatives
		deriv_pcc(0,0) = - deriv_oc(0,0);
		deriv_pcc(1,0) = - deriv_oc(1,0);

		// Yo-derivatives
		deriv_pcc(0,1) = - deriv_oc(0,1);
		deriv_pcc(1,1) = - deriv_oc(1,1);

		// Zo-derivatives
		deriv_pcc(0,2) = - deriv_oc(0,2);
		deriv_pcc(1,2) = - deriv_oc(1,2);
	
		// get derivatives of the rotation parameters
		eo->calc_derivs_rot(deriv_rot, h, u, v, w, d, M, a);
	
		// derivatives of position correction parameters (optional)
		if (_pos_cor != 0)
			_pos_cor->calc_derivs(t, deriv_pcc, deriv_pc);
		
		// derivatives of attitude correction parameters (optional)
		if (_att_cor != 0)
			_att_cor->calc_derivs(t, deriv_rot, deriv_ac);
	
		// derivatives of IMU misalignment parameters (optional)
		if (_imu_mis != 0)
			_imu_mis->calc_derivs(deriv_mis, h, u, v, w, d, R);
	
		// partial derivatives of APs (optional)
		//get_derivs(pho_c, deriv_ap);
		_camera->get_derivs(pho_c, deriv_ap);

		// image coords. derived from the unknown parameters
		DPoint2 img_c0;
		//object_to_image(op->crds, img_c0);
		//projection(pcc, N, op->crds, pho_c);
		pho_c.x = pho_c.z * u / w;
		pho_c.y = pho_c.z * v / w;
		//cout << "pho_c: " << pho_c << endl; cout.flush();
	
		// photo-to-image coords. transformation, including AP correction
		//photo_to_image(pho_c, img_c0);
		_camera->photo_to_image(pho_c, img_c0);
		//cout << "img_c0: " << img_c0 << endl;
		
		// calculate shorten observations (observed - computed term)
		// misclosure vector
		Vector<double> wt(2);
		wt(0) = img_c.x - img_c0.x;
		wt(1) = img_c.y - img_c0.y;
		con_eqn.set_w(wt);
		//cout << "w: " << wt << endl; cout.flush();

		// design matrix B
		con_eqn.set_b(deriv_obs);
		//cout << "B: " << deriv_obs << endl;

		// add contributions to normal equation
		if (!con_eqn.add_contr(atpa, atpl, ltpl))
			return false;
		//con_eqn.print(cout); cout << endl;
		
		num_obs += 2;
		}

	return true;
}


// non-member functions ...

} // namespace DEP
