/**********************************************************************/
//
// FILE:    imumis.cpp
//          IMU misalignment class functions.
// CREATED: 27 Feb 04  (ds)
// CHANGES: see include file
// REFS:    dito 
//
// (c) 2004 Dirk Stallmann (ds)/ IfP Uni Stuttgart
/**********************************************************************/

#include "depvec.h"
#include "imumis.h"
#include "scanner.h"

namespace DEP
{

// constructor

IMUMisalignment::IMUMisalignment(size_t sz) : AdjObject(sz),
_image_ids(),
_a(sz, 0.), _sa(sz, 0.)
{
	_size = sz; // enforce setting number of unknowns???
}


// read

bool IMUMisalignment::read(istream& is)
{
	is >> _id;
	if (!is) return false;
	
	// read image IDs
	string image_id;
	while (is >> image_id && image_id.compare("stop-dep") != 0)
		_image_ids.push_back(image_id);
	
	// rotation matrix flag
	bool matrix_flg = false;
	is >> matrix_flg;
	if (!is) return false;
	
	if (matrix_flg)
		{
		// read rotation matrix
		DRotMatrix3 M;
		is >> M;
		if (!is) return false;
		calc_rot_params(_a, M); // get rotation angles
			
		// read std. devs.
		for (size_t i = 0; i < _sa.size(); i++)
			is >> _sa[i];
		}
	else
		{
		// read rotation parameter and std. devs.
		for (size_t i = 0; i < _a.size(); i++)
			is >> _a[i] >> _sa[i];
		}
	
	_approx_flag = true;
	
	return true;
}


// write

void IMUMisalignment::write(ostream& os) const
{
	Form sci4(4); sci4.scientific().width(12);
	Form sci8(8); sci8.scientific().width(16);
	
	os << key_str() << "   ";
	os << _id << endl;
	/*
	for (size_t i = 0; i < _image_ids.size(); i++)
		{
		os << " " << _image_ids[i];
		if ((i + 1) % 4 == 0) os << "\n";
		}
	os << "stop-dep\n";
	*/
	os << 0 << endl; // matrix flag
	
	for (size_t i = 0; i < _a.size(); i++)
		os << sci8(_a[i]) << "   " << sci4(_sa[i]) << endl;
	
	os << endl;
}


// print

void IMUMisalignment::print(ostream& os) const
{
	os << "IMU misalignment ID      :  " << _id << endl;
	os << "Images using this set    :  " << _image_ids.size();
	{
		for (int i = 0; i < int(_image_ids.size()); i++)
			{
			if (i % 4 == 0) os << endl;
			os << "  " << _image_ids[i];
			}
		os << endl;
	}
	os << "Subtype                  :  " << type_str() << endl;
	//os << "Number of rot. params.   :  " << _a.size() << endl;
	//os << "Rotation matrix form   :  "
	//	<< (_trn_flg ? "transposed" : "normal") << endl;
	os << endl;
	
	Form str4; str4.left().width(4);
	Form str8; str8.left().width(8);
	Form sci8(8); sci8.scientific().width(16);
	Form sci12(3); sci12.scientific().width(12);

	os << "  Param.     State         Value       Std.dev.\n\n";
	for (size_t i = 0; i < _a.size(); i++)
		os << "  " << str8(uname(i)) << "   " << str4(state_str(_sa[i]))
			<< "   " << sci12(_a[i]) << "   " << sci12(_sa[i]) << endl;
	os << endl;
	
	DRotMatrix3 M;
	get_rot_matrix(M);
	Form fix(7); fix.fixed().width(15);
	os << "Rotation matrix:\n";
	os << " " << fix(M(0,0)) << fix(M(0,1)) << fix(M(0,2)) << endl;
	os << " " << fix(M(1,0)) << fix(M(1,1)) << fix(M(1,2)) << endl;
	os << " " << fix(M(2,0)) << fix(M(2,1)) << fix(M(2,2)) << endl;
	os << endl;
}


// get adjustment object class name

string IMUMisalignment::class_str() const
{
	ostringstream oss;
	oss << "IMU misalignment";
	return oss.str();
}


// set upari

bool IMUMisalignment::set_upari()
{
	//cout << _size << endl;
	if (create() != true) return false;

	for (size_t i = 0; i < _a.size(); i++)
		set(i, &_a[i], &_sa[i]);

	return true;
}


// attach to images

bool IMUMisalignment::attach(vector<Image*>& images)
{
	vector<string>::iterator it;
	for (it = _image_ids.begin(); it != _image_ids.end(); ++it)
		{
		Image* p = find(images, *it);
		if (p == 0)
			{
			cout << "WARNING: Image '" << *it
				<< "' for IMU misalign. '"  << _id << "' not found." << endl;
			continue;
			}
			
		ExtFrameImage* img = dynamic_cast<ExtFrameImage*>(p);
		if (img == 0)
			{
			cout << "ERROR: Invalid type of image '" << *it << "'\n";
			return false;
			}
		
		img->attach(this);
		this->inuse(true);
		}

	return true;
}

} // namespace DEP
