/**********************************************************************/
//
// FILE:    imuopkfix.cpp
//          IMU misalignment class functions.
// CREATED: 01 Nov 04  (ds)
// CHANGES: see include file
// REFS:    dito 
//
// (c) 2004 Dirk Stallmann (ds)/ IfP Uni Stuttgart
/**********************************************************************/

#include "imuopkfix.h"

namespace DEP
{

// read

bool IMUMisalignmentOPKFix::read(istream& is)
{
	is >> _id;
	if (!is) return false;
	
	// read image IDs
	string image_id;
	while (is >> image_id && image_id.compare("stop-dep") != 0)
		_image_ids.push_back(image_id);
	
	// rotation matrix flag
	bool matrix_flg = false;
	is >> matrix_flg;
	if (!is)
		{
		cout << "ERROR: Cannot read rotation matrix flag.\n";
		return false;
		}
	
	// unit of rotation angles
	string str;
	is >> str;
	if (!is) return false;
	
	UnitAngle uangle;
	UnitAngle::ecode type = uangle.code(str.c_str());
	if (type == UnitAngle::UNKNOWN) return false;
	const double rho = uangle.rho(type);
	
	if (matrix_flg)
		{
		// read rotation matrix
		DRotMatrix3 M;
		is >> M;
		if (!is) return false;
		calc_rot_params(_a, M); // get rotation angles
		
		// read std. devs.
		is >> _sa[0] >> _sa[1] >> _sa[2];
		}
	else
		{
		// read rotation parameter and std. devs.
		is >> _a[0] >> _sa[0];
		is >> _a[1] >> _sa[1];
		is >> _a[2] >> _sa[2];
		
		// for rotation angles conversion into rad!
		_a[0] /= rho;
		_a[1] /= rho;
		_a[2] /= rho;
		}
	
	// for std. devs. conversion into rad! only for real std. devs.
	if (AdjObject::lspars.is_rsdev(_sa[0])) _sa[0] /= rho;
	if (AdjObject::lspars.is_rsdev(_sa[1])) _sa[1] /= rho;
	if (AdjObject::lspars.is_rsdev(_sa[2])) _sa[2] /= rho;
	
	_approx_flag = true;
	
	return true;
}


// write

void IMUMisalignmentOPKFix::write(ostream& os) const
{
	os << key_str() << "   ";
	os << _id << endl;
	{
		for (size_t i = 0; i < _image_ids.size(); i++)
			{
			os << " " << _image_ids[i];
			if ((i + 1) % 4 == 0) os << "\n";
			}
		os << "stop-dep\n";
	}
	os << 0 << " "; // matrix flag
	
	Form fix4(1); fix4.fixed().width(15);
	Form sci4(4); sci4.scientific().width(15);
	Form sci8(8); sci8.scientific().width(15);
	
	const double rho = AdjObject::lspars.rho();
	os << AdjObject::lspars.u_angle_cstr() << endl;
	
	for (size_t i = 0; i < _a.size(); i++)
		{
		os << sci8(_a[i]) << " ";
		if (is_rsdev(_sa[i]))
			os << fix4(_sa[i] * rho);
		else
			os << sci4(_sa[i]);
		os << endl;
		}
	
	os << endl;
}


// print

void IMUMisalignmentOPKFix::print(ostream& os) const
{
	os << "IMU misalignment ID      :  " << _id << endl;
	os << "Images using this set    :  " << _image_ids.size();
	{
		for (size_t i = 0; i < _image_ids.size(); i++)
			{
			if (i % 4 == 0) os << endl;
			os << "  " << _image_ids[i];
			}
		os << endl;
	}
	os << "Subtype                  :  " << type_str() << endl;
	//os << "Number of rot. params.   :  " << _a.size() << endl;
	//UnitAngle uangle;
	//os << "Unit of rot. angles      :  " << AdjObject::lspars.u_angle_cstr()
	//	<< endl;
	os << endl;
	
	Form str4; str4.left().width(4);
	Form str8; str8.left().width(8);
	Form fix(6); fix.fixed().width(12);
	//Form sci12(3); sci12.scientific().width(12);
	
	const double rho = AdjObject::lspars.rho();
	
	os << "  Param.     State         Value       Std.dev.\n\n";
	for (size_t i = 0; i < _a.size(); i++)
		{
		os << "  " << str8(uname(i)) << "   " << str4(state_str(_sa[i]))
			<< "   " << fix(_a[i] * rho) << "   ";
		print_sdev(os, fix, _sa[i], rho);
// 		if (is_rsdev(_sa[i]))
// 			os << fix9(_sa[i] * rho);
// 		else
// 			os << sci12(_sa[i]);
		os << endl;
		}
	os << endl;
	
	DRotMatrix3 M;
	get_rot_matrix(M);
	fix.precision(7).width(15);
	os << "Rotation matrix:\n";
	os << " " << fix(M(0,0)) << fix(M(0,1)) << fix(M(0,2)) << endl;
	os << " " << fix(M(1,0)) << fix(M(1,1)) << fix(M(1,2)) << endl;
	os << " " << fix(M(2,0)) << fix(M(2,1)) << fix(M(2,2)) << endl;
	os << endl;
}


// get type name

string IMUMisalignmentOPKFix::type_str() const
{
	ostringstream oss;
	oss << "omega-phi-kappa, fixed";
	return oss.str();
}


// get key name

string IMUMisalignmentOPKFix::key_str() const
{
	ostringstream oss;
	oss << "angles-opk-fix";
	return oss.str();
}


// get parameter name

string IMUMisalignmentOPKFix::uname(size_t i) const
{
	ostringstream oss;
	
	switch (i)
		{
		case 0: oss << "omega^b"; break;
		case 1: oss << "phi^b"; break;
		case 2: oss << "kappa^b"; break;
		default:
			assert(false);
			break;
		}
	
	return oss.str();
}


// print changes of unknowns

void IMUMisalignmentOPKFix::print_changes(ostream& os, const Vector<double>& dx) const
{
	Form str4; str4.left().width(4);
	Form str8; str8.left().width(8);
	Form fix(6); fix.fixed().width(12);

	os << "ID         :  " << _id << endl;
	os << "Subtype    :  " << type_str() << endl;
	os << "Index      :  " << _index << endl;
	os << endl;
	
	const double rho = AdjObject::lspars.rho();
	
	os << "  Param.     State"
		"       Approx.         Change       Adjusted\n\n";
	
	size_t i = 0;
	int k = _index;
	
	for (; i < _a.size(); i++, k++)
		{
		os << "  " << str8(uname(i)) << "   "
			<< str4(state_str(i)) << "   "
			<< fix(par(i) * rho) << "   "
			<< fix(dx(k) * rho) << "   "
			<< fix((par(i) + dx(k)) * rho) << endl;
		}
	os << endl;
}


// print unknown parameters

void IMUMisalignmentOPKFix::print_unknowns(ostream& os) const
{
	Form str4; str4.left().width(4);
	Form str8; str8.left().width(8);
	Form fix(6); fix.fixed().width(12);
	Form sci(3); sci.scientific().width(12);

	os << "ID         :  " << _id << endl;
	os << "Subtype    :  " << type_str() << endl;
	os << "Index      :  " << _index << endl;
	os << endl;
	
	const double rho = AdjObject::lspars.rho();
	
	os << "  Param.    State      Initial      Adjusted        Change      "
		"Std.dev.\n\n";

	size_t i = 0;
	
	for (; i < 3; i++)
		{
		os << "  " << str8(uname(i)) << "  " << str4(state_str(i)) << "  "
			<< fix(ival(i) * rho) << "  "
			<< fix(par(i) * rho) << "  "
			<< fix(change(i) * rho) << "  ";
		if (is_rsdev(sdev(i)))
			os << fix(sdev(i) * rho);
		else
			os << sci(sdev(i));
			//print_sdev(os, fix, sdev(i), rho);
		os << endl;
		}
	os << endl;

	DRotMatrix3 M;
	get_rot_matrix(M);
	fix.precision(7);
	os << "Rotation matrix:\n";
	os << " " << fix(M(0,0)) << fix(M(0,1)) << fix(M(0,2)) << endl;
	os << " " << fix(M(1,0)) << fix(M(1,1)) << fix(M(1,2)) << endl;
	os << " " << fix(M(2,0)) << fix(M(2,1)) << fix(M(2,2)) << endl;
	os << endl;
}


// print precision of unknowns

void IMUMisalignmentOPKFix::print_precision(ostream& os) const
{
	os << "ID       :  " << _id << endl;
	os << "Subtype  :  " << type_str() << endl;
	os << "Index    :  " << _index << endl;
	os << endl;

	const double rho = AdjObject::lspars.rho();
	
	os << "  Param.    State      Initial      Adjusted        Change      "
		"Std.dev.    Ratio\n\n";
	
	Form str4; str4.left().width(4);
	Form str8; str8.left().width(8);
	Form fix3(1); fix3.fixed().width(6);
	Form fix6(6); fix6.fixed().width(12);

	for (size_t i = 0; i < _size; i++)
		{
		double t_abs = ratio(i); // test value

		os << "  " << str8(uname(i)) << "  " << str4(state_str(i)) << "  "
			<< fix6(ival(i) * rho) << "  "
			<< fix6(par(i) * rho) << "  "
			<< fix6(change(i) * rho) << "  ";
		//	<< sci11(sdev(i)) << "  " << fix3(t_abs);
		print_sdev(os, fix6, sdev(i), rho);
		os << "  " << fix3(t_abs);

		// test on significance
		if (state(i) == eli)
			os << "  is eliminated";
		else
			{
			if (chk_sign(t_abs))
				os << "  is significant";
			else
				os << "  not significant";
			}
		os << endl;
		}
	os << endl;
}


// calculate derivatives

void IMUMisalignmentOPKFix::calc_derivs(DerivSub& deriv,
double h, double u, double v, double w,
const DPoint3& d, const DRotMatrix3& R) const
{
	deriv.index(_index);
	deriv.newsize(2, 3);
	
	DRotMatrix3 dM_om;
	DRotMatrix3 dM_ph;
	DRotMatrix3 dM_ka;
	dm(dM_om, dM_ph, dM_ka);
	
	DRotMatrix3 dN_om = R * dM_om;
	DRotMatrix3 dN_ph = R * dM_ph;
	DRotMatrix3 dN_ka = R * dM_ka;

	DPoint3 dd_om; mult_inv(dd_om, dN_om, d);
	DPoint3 dd_ph; mult_inv(dd_ph, dN_ph, d);
	DPoint3 dd_ka; mult_inv(dd_ka, dN_ka, d);
	
	deriv(0,0) = h * (w * dd_om.x - u * dd_om.z);
	deriv(1,0) = h * (w * dd_om.y - v * dd_om.z);
	
	deriv(0,1) = h * (w * dd_ph.x - u * dd_ph.z);
	deriv(1,1) = h * (w * dd_ph.y - v * dd_ph.z);
	
	deriv(0,2) = h * (w * dd_ka.x - u * dd_ka.z);
	deriv(1,2) = h * (w * dd_ka.y - v * dd_ka.z);
}

} // namespace DEP
