/**********************************************************************/
//
// FILE:    oridata.cpp
//          Exterior orientation class functions.
// CREATED: 09 Feb 04  (ds)
// CHANGES: see include file
// REFS:    dito 
//
// (c) 2004 Dirk Stallmann (ds)/ IfP Uni Stuttgart
/**********************************************************************/

#include <fstream>
#include "oridata.h"
#include "drvsub.h"
#include "drvrotmat.h"

namespace DEP
{

// constructor

OrientationData::OrientationData() :
	_id(),
	_offset(0),
	_rtype(ERotMatrix3::UNDEFINED),
	_uangle(UnitAngle::NONE),
	_transpose(false),
	_data(),
	_inuse(false)
{
	// empty
}


// read parameters

bool OrientationData::read_params(istream& is)
{
	is >> shift;
	
	string str;
	is >> str;
	
	if (str.compare("trunc") == 0)
		ipol = TRUNC;
	else if (str.compare("linear") == 0)
		ipol = LINEAR;
	else
		return false;
	
	return true;
}


// print parameters

void OrientationData::print_params(ostream& os)
{
	Form fix3(3); fix3.fixed().width(12);
	
	os << "Shift                  :  " << fix3(shift) << " rows\n";
	os << "Interpolation method   :  ";
	switch (ipol)
	{
		case TRUNC: os << "truncate"; break;
		case LINEAR: os << "linear"; break;
		default:
			assert(false);
	}
	os << endl;
	os << endl;
}


// read

bool OrientationData::read(const char* fname)
{
	ifstream is(fname);
	if (!is) return false;
	
	is >> _id;
	if (!is) return false;
	
	// offset
	is >> _offset;
	if (!is) return false;
	
	// type of rotation parameters
	string str;

	is >> str;
	if (!is)
		{
		cout << "OrientationData::read: "
			"Reading type of rotation parameters failed.\n";
		return false;
		}

	_rtype = ERotMatrix3::code(str.c_str());
	if (_rtype == ERotMatrix3::UNDEFINED)
		{
		cout << "OrientationData::read: "
			"Unknown type of rotation parameters: " << str << endl;
		return false;
		}
	//_ = r.size(_rtype);

	// unit of rotation angles
	is >> str;
	if (!is)
		{
		cout << "OrientationData::read: "
			"Reading unit of rotation angles failed.\n";
		return false;
		}

	UnitAngle uangle;
	_uangle = uangle.code(str.c_str());
	if (_uangle == UnitAngle::UNKNOWN)
		{
		cout << "OrientationData::read: "
			"Unknown unit of rotation angles: " << str << endl;
		return false;
		}
	double rho = uangle.rho(_uangle);
	
	// read transpose flag
	is >> _transpose;
	
	// read number of data elements
	size_t sz = 0;
	is >> sz;
	_data = vector<OrientationDataElement3>(sz);
	/*
	if (_data == 0)
		{
		cout << "OrientationData::read: Memory allocation failed.\n";
		return false;
		}
	*/
	
	// number of rotation parameters
	//_num_rot = rtype.size(_rtype);
	//size_t num_rot = ERotMatrix3::size(_rtype);
	
	// read time, position coords. and attitude parameter
	// for rotation angles conversion into rad
	for (size_t i = 0; i < _data.size(); i++)
		{
		is >> _data[i].t;
		is >>_data[i].xo >> _data[i].yo >> _data[i].zo;
		is >>_data[i].a1 >> _data[i].a2 >> _data[i].a3;
			
		_data[i].a1 /= rho;
		_data[i].a2 /= rho;
		_data[i].a3 /= rho;
		}

	// check number of missing rows!!!

	return true;
}


// write

void OrientationData::write(ostream& os) const
{
	os << "orientation-data" << endl;
	os << _id << endl;

	assert(false);
	os << endl;
}


// print

void OrientationData::print(ostream& os) const
{
	Form fix3(3); fix3.fixed().width(12);
	Form fix5(5); fix5.fixed().width(12);
	Form fix9(9); fix9.fixed().width(14);

	os << "ID                     :  " << _id << endl;
	os << "File format            :  " << fmt() << endl;
	os << "Row offset             :  " << _offset << endl;
	os << "Type of rot. params.   :  " << ERotMatrix3::name(_rtype) << endl;
	os << "Use transposed matrix  :  " << (_transpose ? "yes" : "no") << endl;
	os << "Number of rows         :  " << _data.size() << endl;
	os << endl;
	
	double rho = UnitAngle::rho(_uangle);
	
	os << "Start row:\n";
	os << "   t    = " << fix5(_data[0].t)  << " s\n";
	os << "   Xo   = " << fix3(_data[0].xo) << " m\n";
	os << "   Yo   = " << fix3(_data[0].yo) << " m\n";
	os << "   Zo   = " << fix3(_data[0].zo) << " m\n";
	os << "   a1 = " << fix9(_data[0].a1 * rho) << " " << UnitAngle::symbol(_uangle) << endl;
	os << "   a2 = " << fix9(_data[0].a2 * rho) << " " << UnitAngle::symbol(_uangle) << endl;
	os << "   a3 = " << fix9(_data[0].a3 * rho) << " " << UnitAngle::symbol(_uangle) << endl;
	os << endl;

	int l = _data.size() - 1;
	os << "Last row:\n";
	os << "   t    = " << fix5(_data[l].t)  << " s\n";
	os << "   Xo   = " << fix3(_data[l].xo) << " m\n";
	os << "   Yo   = " << fix3(_data[l].yo) << " m\n";
	os << "   Zo   = " << fix3(_data[l].zo) << " m\n";
	os << "   a1 = " << fix9(_data[l].a1 * rho) << " " << UnitAngle::symbol(_uangle) << endl;
	os << "   a2 = " << fix9(_data[l].a2 * rho) << " " << UnitAngle::symbol(_uangle) << endl;
	os << "   a3 = " << fix9(_data[l].a3 * rho) << " " << UnitAngle::symbol(_uangle) << endl;
	os << endl;
	os << endl;
}


// get file format

string OrientationData::fmt() const
{
	ostringstream oss;
	oss << "DEP";
	return oss.str();
}


// get time interval

void OrientationData::time_interval(double& t1, double& tn) const
{
	t1 = _data[0].t;
	tn = _data[_data.size() - 1].t;
}


// get time

double OrientationData::get_time(int j) const
{
	int i = j - _offset;
	
	return (_data[i].t);
}


// get position coordinates

void OrientationData::get_per_centre(int j, DPoint3& pos) const
{
	int i = j - _offset;

	pos.x = _data[i].xo;
	pos.y = _data[i].yo;
	pos.z = _data[i].zo;
}


// get rotation parameters

void OrientationData::get_rot_params(int j, vector<double>& a) const
{
	int i = j - _offset;
	
	a[0] = _data[i].a1;
	a[1] = _data[i].a2;
	a[2] = _data[i].a3;
}


// get rotation matrix

void OrientationData::get_rot_matrix(int j, DRotMatrix3& rm) const
{	
	vector<double> a(3);
	get_rot_params(j, a);
	calc_rot_matrix(rm, a);
}


// calculate position difference

void OrientationData::pos_difference(int j, DPoint3& d) const
{
	int i = j - _offset;
	//if (!inside(i)) return false;

	int k = i + 1;
	//if (!inside(j)) return false;

	d.x = _data[k].xo - _data[i].xo;
	d.y = _data[k].yo - _data[i].yo;
	d.z = _data[k].zo - _data[i].zo;
}


// calculate interpolated time

double OrientationData::get_time_trunc(double r) const
{
	int i = int(r - shift) - _offset;
	
	return (_data[i].t);
}


// calculate interpolated perspective centre coordinates

void OrientationData::get_per_centre_trunc(double r, DPoint3& pos) const
{
	int i = int(r - shift) - _offset;
	
	pos.x = _data[i].xo;
	pos.y = _data[i].yo;
	pos.z = _data[i].zo;
}


// calculate interpolated rotations

void OrientationData::get_rot_params_trunc(double r, vector<double>& a) const
{
	int i = int(r - shift) - _offset;
	
	a[0] = _data[i].a1;
	a[1] = _data[i].a2;
	a[2] = _data[i].a3;
}


// calculate interpolated rotation matrix

void OrientationData::get_rot_matrix_trunc(double r, DRotMatrix3& rm) const
{	
	vector<double> a(3);
	get_rot_params_trunc(r, a);
	calc_rot_matrix(rm, a);
}


// calculate interpolated time

double OrientationData::get_time_linear(double r) const
{
	int i = int(r - shift) - _offset;
	
	double dt = _data[i+1].t - _data[i].t;
	double a = r + double(_offset) - double(i);
	
	return (_data[i].t + a * dt);
}


// calculate interpolated perspective centre coordinates

void OrientationData::get_per_centre_linear(double r, DPoint3& pos) const
{
	int i = int(r - shift) - _offset;
	
	double m = r - shift - double(i + _offset);
	
	pos.x = _data[i].xo + m * (_data[i+1].xo - _data[i].xo);
	pos.y = _data[i].yo + m * (_data[i+1].yo - _data[i].yo);
	pos.z = _data[i].zo + m * (_data[i+1].zo - _data[i].zo);
}


// calculate interpolated rotations

void OrientationData::get_rot_params_linear(double r, vector<double>& a) const
{
	int i = int(r - shift) - _offset;
	
	double m = r - shift - double(i + _offset);
	
	a[0] = _data[i].a1 + m * (_data[i+1].a1 - _data[i].a1);
	a[1] = _data[i].a2 + m * (_data[i+1].a2 - _data[i].a2);
	a[2] = _data[i].a3 + m * (_data[i+1].a3 - _data[i].a3);
}


// calculate interpolated rotation matrix

void OrientationData::get_rot_matrix_linear(double r, DRotMatrix3& rm) const
{	
/*
	DRotMatrix3 R0, R1, dR;
	R0.set(_data[i].a, _rtype);
	R1.set(_data[i+1].a, _rtype);
	R0.transpose();
	dR = R0 * R1;
	
	double eps_x = dR(2,1);
	double eps_y = dR(0,3);
	double eps_z = dR(1,0);
	cout << eps_x << " " << eps_y << " " << eps_z << endl;
	
	double a = r + double(_offset - i);
	cout << a << endl;
	eps_x *= a;
	eps_y *= a;
	eps_z *= a;
	cout << eps_x << " " << eps_y << " " << eps_z << endl;
	
	dR.set_angles_small(eps_x, eps_y, eps_z);
	
	R0.set(_data[i].a, _rtype);
	rm = R0 * dR;
	*/
	
	vector<double> a(3);
	get_rot_params_linear(r, a);
	calc_rot_matrix(rm, a);
}


// calculate interpolated time

double OrientationData::get_time(double r) const
{
	double t = 0;
	
	switch (ipol)
		{
		case TRUNC:
			t = get_time_trunc(r);
			break;
		
		case LINEAR:
			t = get_time_linear(r);
			break;
		
		default:
			assert(false);
		}
	
	return t;
}


// calculate interpolated perspective centre coordinates

void OrientationData::get_per_centre(double r, DPoint3& pos) const
{
	switch (ipol)
		{
		case TRUNC:
			get_per_centre_trunc(r, pos);
			break;
		
		case LINEAR:
			get_per_centre_linear(r, pos);
			break;
		
		default:
			assert(false);
		}
}


// calculate interpolated rotations

void OrientationData::get_rot_params(double r, vector<double>& a) const
{
	switch (ipol)
		{
		case TRUNC:
			get_rot_params_trunc(r, a);
			break;
		
		case LINEAR:
			get_rot_params_linear(r, a);
			break;
		
		default:
			assert(false);
		}
}


// calculate interpolated rotation matrix

void OrientationData::get_rot_matrix(double r, DRotMatrix3& rm) const
{
	switch (ipol)
		{
		case TRUNC:
			get_rot_matrix_trunc(r, rm);
			break;
		
		case LINEAR:
			get_rot_matrix_linear(r, rm);
			break;
		
		default:
			assert(false);
		}
}

} // namespace DEP
