/**********************************************************************/
//
// FILE:    oridata.h
//          Orientation data class.
// CREATED: 09 Feb 04  (ds)
// CHANGES: 
// REFS:
//
// (c) 2004 Dirk Stallmann (ds)/ IfP Uni Stuttgart
/**********************************************************************/

#ifndef DEP_ORIDATA_INCLUDED
#define DEP_ORIDATA_INCLUDED

#include "depdef.h"

using namespace std;
using namespace GenLib2;

namespace DEP
{
class DerivSub;

/** Orientation data element
 *  A single motion data element consist of the time stamp, position and
 *  orientation.
 *  The position is described by a vector (xo, yo, yo)
 *  while the orientation is represented variable by 3 (a1, a2, a3)
 *  rotation parameters.
 *
 *  @author Dirk Stallmann
 */
struct OrientationDataElement3
{
public:	
	/// corresponding time
	double		t;

	/// perspective centre coordinates
	double		xo;
	double		yo;
	double		zo;

	/// 3 rotation parameters
	double		a1;
	double		a2;
	double		a3;
};


/** Orientation data set class.
 *  This class holds the exterior orientation data for each scanline
 *  from a pushbroom scanner.
 *
 *  @author Dirk Stallmann
 */
class OrientationData
{
public:
	/// interpolation method
	enum IpolMethod
	{
		TRUNC,	// truncate
		LINEAR	// linear
	};

	/// constructor
	OrientationData();

	/// destructor
	virtual ~OrientationData() {}
	
	/**@name read/print parameters */
	//@{
		/// read parameters
		static bool read_params(istream& is);
		
		/// print parameters
		static void print_params(ostream& os);
	//@}
	
	/**@name read/write/print orientation data set */
	//@{
		/// read
		virtual bool read(const char* fname);

		/// write
		virtual void write(ostream& os) const;

		/// print
		virtual void print(ostream& os) const;
	//@}
	
	/**@name set/get member data */
	//@{
		/// set data set ID
		void set_id(const char* id) { _id = id; }
	
		/// get data set ID
		string get_id() const { return _id; }

		/// get file format
		virtual string fmt() const;
		
		/// get row offset
		int get_offset() const { return _offset; }

		/// number of rotation parameters
		size_t num_rot_params() const
		{
			return ERotMatrix3::size(_rtype);
		}
		
		/// get number of rows
		size_t nrows() const { return _data.size(); }
		
		/** get time interval,
		 *  get time of first and last scanline
		 */
		void time_interval(double& t1, double& tn) const;
	//@}
	
	/// check scanline number if inside data range [0, nrows]
	bool inside(int j) const
	{
		int i = int(j) - _offset;
		return (i >= 0 && i < int(_data.size()) - 1);
	}
	
	/** check scanline number
	 *  Check if inside data range [-0.5, (nrows-0.5)].
	 *  Asuming the fractional part is equal zero at the
	 *  centre of the scanline.
	 */
	bool inside(double r) const
	{
		double i = r - double(_offset) - shift;
		return (i >= 0.5 && i < double(_data.size()) - 0.5);
	}
	
	/** get orientation data by index j
	  * Asuming integer row offset but ignore global shift.
	  * Without range-checking! For test use inside() method.
	  */
	//@{
		/// get time
		double get_time(int j) const;

		/// get perspective centre coordinates
		void get_per_centre(int j, DPoint3& pos) const;

		/// get rotation parameters
		void get_rot_params(int j, vector<double>& _rot_params) const;
		
		/// get rotation matrix
		virtual void get_rot_matrix(int j, DRotMatrix3& rm) const;
		
		/// calculate position difference between
		void pos_difference(int j, DPoint3& d) const;
	//@}
	
	/** calculate orientation data interpolation
	  * asuming offset and without range-checking
	  */
	//@{
		/// get interpolated time
		double get_time_trunc(double r) const;
		
			/// get interpolated perspective centre coordinates
		void get_per_centre_trunc(double r, DPoint3& pos) const;
		
			/// get interpolated rotations
		void get_rot_params_trunc(double r, vector<double>& a) const;
		
			/// get interpolated rotation matrix
		void get_rot_matrix_trunc(double r, DRotMatrix3& rm) const;
	//@}
	
	/** calculate orientation data by lineare interpolation
	  * asuming offset and without range-checking
	  */
	//@{
		/// calculate interpolated time
		double get_time_linear(double r) const;

		/// calculate interpolated perspective centre coordinates
		void get_per_centre_linear(double r, DPoint3& pos) const;

		/// calculate interpolated rotations
		void get_rot_params_linear(double r, vector<double>& a) const;
		
		/// calculate interpolated rotation matrix
		void get_rot_matrix_linear(double r, DRotMatrix3& rm) const;
	//@}
	
	/** calculate orientation data by lineare interpolation
	  * Asuming offset and global shift.
	  * Without range-checking. For test use inside() method!
	  */
	//@{
		/// calculate interpolated time
		double get_time(double r) const;
	
		/// calculate interpolated perspective centre coordinates
		void get_per_centre(double r, DPoint3& pos) const;
	
		/// calculate interpolated rotations
		void get_rot_params(double r, vector<double>& a) const;
	
		/// calculate interpolated rotation matrix
		void get_rot_matrix(double r, DRotMatrix3& rm) const;
	//@}
	
	/// set inuse flag
	void inuse(bool flg) { _inuse = flg; }

	/// test inuse flag
	bool is_unused() const { return _inuse == false; }
	
	/// calculate rotation matrix from rotation parameters
	void calc_rot_matrix(DRotMatrix3& rm, const vector<double>& a) const
	{
		rm.set(a, _rtype);
		if (_transpose) rm.transpose();
	}
	
	// calculate derivatives of rotation parameters
	//void calc_derivs_rot(DerivSub& deriv,
	//	double h, double u, double v, double w,
	//	const DPoint3& d, const DRotMatrix3& M, const vector<double>& a) const;

protected:
	/// ID
	string		_id;

	/// row/line offset
	int			_offset;

	/// type of rotation parameters
	ERotMatrix3::ecode	_rtype;

	/// unit of rotation angles
	UnitAngle::ecode	_uangle;

	/// transpose flag; transpose matrix after calculation
	bool		_transpose;
	
	/// orientation data elements
	vector<OrientationDataElement3>	_data;

	/// inuse flag
	bool		_inuse;

public:
	/// global row/line shift
	static double shift;

	/// interpolation method
	static IpolMethod ipol;
};


// non-member functions ...

} // namespace DEP

#endif
