/**********************************************************************/
//
// FILE:    scanner.cpp
//          Scanner image class functions.
// CREATED: 24 March 05  (ds)
// CHANGES: see include file
// REFS:    dito 
//
// (c) 2005 Dirk Stallmann (ds)/ IfP Uni Stuttgart
/**********************************************************************/

#include "depvec.h"
#include "scanner.h"
#include "drvsub.h"
#include "lconeqn.h"
#include "imgpnt.h"
#include "objpnt.h"
#include "ccdcam.h"
#include "oridata.h"
#include "extori.h"
#include "poscor.h"
#include "attcor.h"
#include "imumis.h"

namespace DEP
{

// constructor

ScannerImage::ScannerImage() :
	ExtFrameImage(),
	_ods_id(),
	_ods(0)
{
	// empty
}


// read image configuration

bool ScannerImage::read(istream& is, const vector<Camera*>& cams)
{
	// read image, ODS and camera ID
	is >> _image_id >> _ods_id >> _camera_id;
	if (!is) return false;

	// attach camera
	assert(_camera == 0); // assume camera is not defined!

	if (!Image::attach(cams))
		{
		cout << "ERROR: Attach camera " << _camera_id
			<< " for image " << _image_id << " failed.\n";
		return false;
		}

	return true;
}


// print image configuration

void ScannerImage::print(ostream& os) const
{
	os << "Image ID                :  " << _image_id << endl;
	os << "Image type              :  " << type_str() << endl;
	os << "Orientation data set ID :  " << _ods_id << endl;
	os << "Camera ID               :  " << _camera_id << endl;
	os << endl;
}


// get image type

string ScannerImage::type_str() const
{
	ostringstream oss;
	oss << "scanner";
	return oss.str();
}


// attach orientation data set

bool ScannerImage::attach(const vector<OrientationData*>& ods)
{
	_ods = find(ods, _ods_id);
	if (_ods == 0)
		{
		cout << "ERROR: Attach orientation data set " << _ods_id
			<< " for image " << _image_id << " failed.\n";
		return false;
		}
		
	_ods->inuse(true);
	
	return true;
}


// calculate exterior orientation

bool ScannerImage::calc_ext_orientation(DPoint3& pcc, DRotMatrix3& N, double r) const
{
	assert(_ods != 0);
	
	// range check
	if (_ods->inside(r) != true) return false;
	
	// instantaneous perspective centre coords.
	_ods->get_per_centre(r, pcc);
	
	// instantaneous time
	double t = _ods->get_time(r);
	
	// position correction (optional)
	if (_pos_cor != 0)
		pcc += _pos_cor->calc_corr(t);
	
	// rotation parameters
	size_t n = _ods->num_rot_params(); // test missing!
	vector<double> a(n);
	_ods->get_rot_params(r, a);
	
	// attitude correction (optional)
	if (_att_cor != 0)
		{
		DPoint3 da = _att_cor->calc_corr(t);
		a[0] += da.x;
		a[1] += da.y;
		a[2] += da.z;
		}
		
	// instantaneous rotation matrix
	DRotMatrix3 R;
	_ods->calc_rot_matrix(R, a);
	
	// IMU misalignment (optional)
	DRotMatrix3 M;
 	if (_imu_mis != 0)
 		_imu_mis->get_rot_matrix(M);
	
	// total rotation matrix
	// attitude correction and misalignment included
	N = R * M;
	
	return true;
}


// object-to-photo coords. transformation

bool ScannerImage::object_to_photo(const DPoint3& obj_c, DPoint3& pho_c, double r) const
{
	assert(_camera != 0);

	// instantaneous perspective centre coords. and rotation matrix
	DPoint3 pcc;
	DRotMatrix3 N;
	
	if (calc_ext_orientation(pcc, N, r) != true)
		return false;
	
	// set principal distance
	pho_c.z = -_camera->ck();
	
	// back projection
	return projection(pcc, N, obj_c, pho_c);
}


// object-to-image coords. transformation

bool ScannerImage::object_to_image(const DPoint3& obj_c, DPoint2& img_c, double r) const
{
	// object-to-photo coords. transformation
	DPoint3 pho_c;
	object_to_photo(obj_c, pho_c, r);
 
	// photo-to-image coords. transformation, including AP correction
	photo_to_image(pho_c, img_c);
	
	return true;
}


// spatial intersection

bool ScannerImage::intersection(const DPoint2& ext_c1, const Image& image2,
const DPoint2& ext_c2, DPoint3& obj_c, double& e) const
{
	const ScannerImage& s2 = dynamic_cast<const ScannerImage&>(image2);
	
// 	// external-to-image coords. transformation
// 	DPoint2 img_c1, img_c2;
// 	this->_camera->external_to_image(ext_c1, img_c1);
// 	s2._camera->external_to_image(ext_c2, img_c2);
// 	
// 	// image-to-photo coords. transformation, AP included
	DPoint3 pho_c1, pho_c2;
// 	this->image_to_photo(img_c1, pho_c1);
// 	s2.image_to_photo(img_c2, pho_c2);
	
	this->sensor_to_photo(ext_c1, pho_c1);
	s2.sensor_to_photo(ext_c2, pho_c2);
	
	// get perspective centre coords. and rotation matrix
	DPoint3 pcc1, pcc2;
	DRotMatrix3 rm1, rm2;
	if (this->calc_ext_orientation(pcc1, rm1, ext_c1.y) != true)
		return false;
	if (s2.calc_ext_orientation(pcc2, rm2, ext_c2.y) != true)
		return false;
/*
	cout << "ext_c1: " << ext_c1 << endl;
	cout << "img_c1: " << img_c1 << endl;
	cout << "pho_c1: " << pho_c1 << endl;
	cout << "pcc1: " << pcc1 << endl;
	cout << "rm1: " << rm1 << endl;
	
	cout << "ext_c2: " << ext_c2 << endl;
	cout << "img_c2: " << img_c2 << endl;
	cout << "pho_c2: " << pho_c2 << endl;
	cout << "pcc2: " << pcc2<< endl;
	cout << "rm2: " << rm2 << endl;
*/
	// do spatial intersection
	return DEP::intersection(pcc1, rm1, pho_c1, pcc2, rm2, pho_c2, obj_c, e);
}


// 

bool ScannerImage::observ_eqn_intersec(const string& id, DerivSub& a, DVector& wt,
bool& found) const
{
	ImagePoint* ip = get_image_point(id);
	if (ip == 0)
		{
		found = false;
		return true; // escape, point not found
		}
	found = true;

	DPoint3 pho_c;
	sensor_to_photo(ip->crds, pho_c);
	
	DPoint3 pcc;
	DRotMatrix3 R;
	
	if (calc_ext_orientation(pcc, R, ip->crds.y) != true)
		{
		cout << "ERROR: ScannerImage::observ_eqn_intersec: Zero denominator.\n";
		return false;
		}
		
	double u = R(0,0) * pho_c.x + R(0,1) * pho_c.y + R(0,2) * pho_c.z;
	double v = R(1,0) * pho_c.x + R(1,1) * pho_c.y + R(1,2) * pho_c.z;
	double w = R(2,0) * pho_c.x + R(2,1) * pho_c.y + R(2,2) * pho_c.z;
	if (fabs(w) == 0.)
		{
		cout << "ERROR: ScannerImage::observ_eqn_intersec: Zero denominator.\n";
		return false;
		}
		
	double ax = - u / w;
	double ay = - v / w;
	
	double bx = - (ax * pcc.z + pcc.x);
	double by = - (ay * pcc.z + pcc.y);
	
	a(0,0) = 1.;
	a(0,1) = 0.;
	a(0,2) = ax;
	
	a(1,0) = 0.;
	a(1,1) = 1.;
	a(1,2) = ay;
	
	wt(0) = -bx;
	wt(1) = -by;

	return true;
}


// add contribution to normal equations for all image point observations
// using the Gauss-Helmert model

bool ScannerImage::normal_eqns(const vector<ObjectPoint*>& obpts,
Matrix<double>& atpa, Vector<double>& atpl, double& ltpl, int& num_obs) const
{
	assert(_ods != 0);		// necessary
	assert(_camera != 0);	// necessary

	// Setup sub-matrices of derivatives
	// Always 2 observation equations for each image point.
	// The number of columns might be variable.
	DerivSub	deriv_oc(2,3);		// object points, size always 2x3
	DerivSub	deriv_pcc(2,3);		// EO params., size always 2x3
	DerivSub	deriv_rot(2,3);	// rotation params., variable size
	vector<DerivSub> deriv_ap(_camera->num_ap_sets()); // APs, variable size
	DerivSub	deriv_pc;			// position corr., variable size
	DerivSub	deriv_ac;			// attitude corr., variable size
	DerivSub	deriv_mis;			// IMU misalignment, variable size
	DerivSub	deriv_obs(2,2);		// observations, size always 2x2
	
	// derivatives of observations
	deriv_obs.identity();

	// Always 2 condition equations for a single imaging ray.
	LinConditionEqn con_eqn;

	// design matrix A
	con_eqn.add_a(&deriv_oc);	// always
	con_eqn.add_a(deriv_ap);	// optional
	if (_pos_cor != 0) con_eqn.add_a(&deriv_pc);	//optional
	if (_att_cor != 0) con_eqn.add_a(&deriv_ac);	//optional
	if (_imu_mis != 0) con_eqn.add_a(&deriv_mis);	//optional
	
	DRotMatrix3 R; // instantenous rotation matrix
	DRotMatrix3 M; // IMU misalignment matrix
	DRotMatrix3 N; // instantenous rotation matrix: N = R M
	
	// get IMU misalignment matrix
	if (_imu_mis != 0) _imu_mis->get_rot_matrix(M);
	
	// loop over all image points
	vector<ImagePoint*>::const_iterator ip;
	for (ip = _image_points.begin(); ip != _image_points.end(); ++ip)
		{
		if (*ip == 0) continue;
		
		// search for corresponding object point
		ObjectPoint* op = find(obpts, (*ip)->get_id());
		if (op == 0)
			{
			cout << "ATTENTION: Object point " << (*ip)->get_id()
				<< " not found.\n";
			continue;
			}

		//cout << _image_id << ":" << (*ip)->get_id() << ":\n";

		// set current index of object point
		deriv_oc.index(op->get_index());

		// convert sensor to image std. dev.
		DPoint2 sdev;
		_camera->sdev_sensor_to_image((*ip)->sdev, sdev);
		//cout << (*ip)->id() << ": " << sdev << endl;

		// co-factors
		Matrix<double> Qll(2,2);
		Qll(0,0) = sdev.x / AdjObject::lspars.sigma0(); Qll(0,0) *= Qll(0,0);
		Qll(1,1) = sdev.y / AdjObject::lspars.sigma0(); Qll(1,1) *= Qll(1,1);
		con_eqn.set_q(Qll);

		// convert sensor into image coords.
		DPoint2 img_c;
		_camera->sensor_to_image((*ip)->crds, img_c);
		//cout << "img: " << img_c << endl;

		// convert image into photo coords., AP included
		DPoint3 pho_c;
		image_to_photo(img_c, pho_c);
		//cout << "pho: " << pho_c << endl;
		
		// row/scanline and range check
		double r = (*ip)->crds.y;
		//cout << "r=" << r << endl;
		if (_ods->inside(r) != true) return false;
		
		// instantaneous time
		double t = _ods->get_time(r);
		
		// instantaneous perspective centre coords.
		DPoint3 pcc;
		_ods->get_per_centre(r, pcc);
	
		// position correction (optional)
		if (_pos_cor != 0)
			pcc += _pos_cor->calc_corr(t);
		//cout << pcc << endl;
		
		// instantaneous rotations
		size_t n = _ods->num_rot_params();
		vector<double> a(n);
		_ods->get_rot_params(r, a);
		
		// attitude correction (optional)
		if (_att_cor != 0)
			{
			DPoint3 da = _att_cor->calc_corr(t);
			a[0] += da.x;
			a[1] += da.y;
			a[2] += da.z;
			}
		
		// instantaneous rotation matrix
		_ods->calc_rot_matrix(R, a);
		//cout << R << endl;
		
		// total rotation matrix
		N = R * M;
		
		// object coords. reduced to the perspective centre
		DPoint3 d;
		d.x = op->crds.x - pcc.x;
		d.y = op->crds.y - pcc.y;
		d.z = op->crds.z - pcc.z;

		double w  = N(0,2) * d.x + N(1,2) * d.y + N(2,2) * d.z;
		if (fabs(w) == 0.)
			{
			cout << "ERROR: ScannerImage::normal_eqns: Zero denominator.\n";
			return false;
			}

		double u = N(0,0) * d.x + N(1,0) * d.y + N(2,0) * d.z;
		double v = N(0,1) * d.x + N(1,1) * d.y + N(2,1) * d.z;

		double g = pho_c.z / w;
		double h = g / w;
	
		// derivatives of the object coordinates (always)
		// X-derivatives
		deriv_oc(0,0) = h * (N(0,0) * w - N(0,2) * u);
		deriv_oc(1,0) = h * (N(0,1) * w - N(0,2) * v);

		// Y-derivatives
		deriv_oc(0,1) = h * (N(1,0) * w - N(1,2) * u);
		deriv_oc(1,1) = h * (N(1,1) * w - N(1,2) * v);

		// Z-derivatives
		deriv_oc(0,2) = h * (N(2,0) * w - N(2,2) * u);
		deriv_oc(1,2) = h * (N(2,1) * w - N(2,2) * v);
	
		// The derivatives of the projection centre coordinates are simply
		// the negative derivatives of the object coordinates !
		// Xo-derivatives
		deriv_pcc(0,0) = - deriv_oc(0,0);
		deriv_pcc(1,0) = - deriv_oc(1,0);

		// Yo-derivatives
		deriv_pcc(0,1) = - deriv_oc(0,1);
		deriv_pcc(1,1) = - deriv_oc(1,1);

		// Zo-derivatives
		deriv_pcc(0,2) = - deriv_oc(0,2);
		deriv_pcc(1,2) = - deriv_oc(1,2);
		
		// derivatives of the rotation parameters
		//_ods->calc_derivs_rot(deriv_rot, h, u, v, w, d, M, a);
	
		{
		DRotMatrix3 dR_a1;
		DRotMatrix3 dR_a2;
		DRotMatrix3 dR_a3;
	
		// call virtial function to get derivatives
		//_ods->calc_derivs_rot(a, deriv_rot);
		derivs_opk_fix_cw(a[0], a[1], a[2], dR_a1, dR_a2, dR_a3);
		dR_a1.transpose();
		dR_a2.transpose();
		dR_a3.transpose();
	
		DRotMatrix3 dN_a1 = dR_a1 * M;
		DRotMatrix3 dN_a2 = dR_a2 * M;
		DRotMatrix3 dN_a3 = dR_a3 * M;
	
		DPoint3 dd_a1; mult_inv(dd_a1, dN_a1, d);
		DPoint3 dd_a2; mult_inv(dd_a2, dN_a2, d);
		DPoint3 dd_a3; mult_inv(dd_a3, dN_a3, d);
	
		deriv_rot(0,0) = h * (w * dd_a1.x - u * dd_a1.z);
		deriv_rot(1,0) = h * (w * dd_a1.y - v * dd_a1.z);
	
		deriv_rot(0,1) = h * (w * dd_a2.x - u * dd_a2.z);
		deriv_rot(1,1) = h * (w * dd_a2.y - v * dd_a2.z);
	
		deriv_rot(0,2) = h * (w * dd_a3.x - u * dd_a3.z);
		deriv_rot(1,2) = h * (w * dd_a3.y - v * dd_a3.z);
		}
		
		
		// derivatives of position correction (optional)
		if (_pos_cor != 0)
			_pos_cor->calc_derivs(t, deriv_pcc, deriv_pc);
		
		// derivatives of attitude correction parameters (optional)
		if (_att_cor != 0)
			_att_cor->calc_derivs(t, deriv_rot, deriv_ac);
		
		// get derivatives of IMU misalignment parameters (optional)
		if (_imu_mis != 0)
			_imu_mis->calc_derivs(deriv_mis, h, u, v, w, d, R);
		
		// partial derivatives of APs (optional)
		//get_derivs(pho_c, deriv_ap);
		_camera->get_derivs(pho_c, deriv_ap);
		
		// image coords. derived from the unknown parameters
		DPoint2 img_c0;
		//object_to_image(r, op->crds, img_c0, r);
		//projection(pcc, N, op->crds, pho_c);
		pho_c.x = pho_c.z * u / w;
		pho_c.y = pho_c.z * v / w;
		//cout << "pho_c0: " << pho_c << endl;
		
		// photo-to-image coords. transformation
		_camera->photo_to_image(pho_c, img_c0);
		//cout << "img_c0: " << img_c0 << endl;
		
		// APs correction
		//img_c0 += ap_set_corr(pho_c);
		
		// calculate shorten observations (observed - computed term)
		// misclosure vector
		Vector<double> wt(2);
		wt(0) = img_c.x - img_c0.x;
		wt(1) = img_c.y - img_c0.y;
		con_eqn.set_w(wt);
		//cout << "w: " << wt << endl;
		
		// design matrix B
		con_eqn.set_b(deriv_obs);
		//cout << "B: " << deriv_obs << endl;

		// add contributions to normal equation
		if (!con_eqn.add_contr(atpa, atpl, ltpl))
			return true;

		num_obs += 2;
		}
	
	return true;
}


// print residual image coords.
/*
void ScannerImage::print_residuals(const vector<ObjectPoint*>& obj_pnts,
DPoint2& sum_v, DPoint2& sum_vv, DPoint2& sum_pvv,
DPoint2& sum_p, int& num, ostream& os) const
{
	assert(_camera != 0);

	Form str8; str8.width(8);
	Form fix10(2); fix10.fixed().width(10);

	DPoint2 img_c0;
	DPoint2 v, vv;
	DPoint2 p;
	double v_abs = 0.;

	// initialize accumulator
	sum_v.x = sum_v.y = 0.;
	sum_vv.x = sum_vv.y = 0.;
	num = 0;
	
    os << "Image ID : " << _image_id << endl << endl;
	os << "  Point ID       vx/um      vy/um     |v|/um\n\n";

	// loop over all image points
	vector<ImagePoint*>::const_iterator ip;
	for (ip = _image_points.begin(); ip != _image_points.end(); ++ip)
		{
		if (*ip == 0) continue;

		// find corresponding object point
		ObjectPoint* op = find(obj_pnts, (*ip)->get_id());
		if (op == 0)
			{
			os << "ATTENTION: Object point "
				<< (*ip)->get_id() << " not found" << endl;
			continue;
			}

		// collinear equation -> calculate image coords.
		// Attention: The scanline coordinate is only used by scanner images!!!
		DPoint2 img_c0;
		object_to_image(op->crds, img_c0, (*ip)->crds.y);

		// convert external into image coords.
		DPoint2 img_c;
		_camera->external_to_image((*ip)->crds, img_c);

		// residuals: Soll - Ist
		v = img_c0 - img_c;
		v_abs = abs(v);
		vv = v * v;

		// convert external into image std. dev.
		DPoint2 sdev;
		_camera->sdev_external_to_image((*ip)->sdev, sdev);

		// weights
		p.x = AdjObject::lspars.sigma0() / sdev.x;
		p.y = AdjObject::lspars.sigma0() / sdev.y;
		p *= p;
		
		sum_v += v;
		sum_vv += vv;
		sum_pvv += vv * p;
		sum_p += p;
		num++;

		// print residual
		if (AdjObject::lspars.is_small(v.x) || AdjObject::lspars.is_small(v.y))
			{
			os << "  " << str8((*ip)->get_id()) << "  "
				<< fix10(v.x * 1e6) << " " << fix10(v.y * 1e6) << " "
				<< fix10(v_abs * 1e6);

			if (op->is_ctrl())
				{
				os << "   control point for";
				if (op->is_csx()) os << " X";
				if (op->is_csy()) os << " Y";
				if (op->is_csz()) os << " Z";
				}
			os << endl;
			}
		}
	os << endl;

	os << "Number of points       :   " << num << endl;
	os << "RMS of residuals       :   "
		<< fix10(sqrt(sum_vv.x / num) * 1e6) << "   "
		<< fix10(sqrt(sum_vv.y / num) * 1e6) << endl;
	os << "Average of residuals   :   "
		<< fix10(sum_v.x * 1e6 / num) << "   "
		<< fix10(sum_v.y * 1e6 / num) << endl;
	os << endl;
}
*/


// non-member functions ...

// ray-z-plane intersection

bool intersection(const CCDCamera& cam, const OrientationData& eo,
const DPoint2& pix_c, DPoint3& obj_c)
{
	DPoint2 img_c;
	DPoint3 pho_c(0., 0., -cam.ck());
	DPoint3 pcc;
	DRotMatrix3 rm;

	// calculate image coords.
	cam.sensor_to_image(pix_c, img_c);
	//cout << "img_c = " << img_c << endl;

	// calculate photo coords.
	cam.image_to_photo(img_c, pho_c);
	//cout << "pho_c = " << pho_c << endl;
	
	// check range
	if (eo.inside(pix_c.y) != true)
		{
		cout << "ERROR: intersection: Index out of EO data range.\n";
		return false;
		}
		
	// get interpolated position and attitude
	eo.get_per_centre(pix_c.y, pcc);
	eo.get_rot_matrix(pix_c.y, rm);

	// ray-z-plane intersection
	intersection(pcc, rm, obj_c, pho_c);
	//cout << "obj_c = " << obj_c << endl;

	return true;
}


// back projection by bisection method

bool projection_bisection(const CCDCamera& cam, const OrientationData& eo,
const DPoint3& obj_c, DPoint& pix_c,
int& nit, bool dbg)
{
	DPoint2 img_c, cal_c;
	DPoint3 pho_c(0., 0., -cam.ck());
	DPoint3 pcc;
	DRotMatrix3 rm;

	int jl = 0;
	int ju = eo.nrows() - 1;

	// iteration loop
	nit = 0;
	while (ju - jl > 1)
		{
		//if (dbg) cout << endl << "Iteration: " << nit << endl;
		//if (dbg) cout << "pix_c = " << pix_c << endl;
		int jm = (ju + jl) >> 1;
		
		// check range
		if (eo.inside(pix_c.y) != true)
			{
			cout << "ERROR: projection_bisection: Index out of EO data range.\n";
			return false;
			}
		
		// get position and attitude
		eo.get_per_centre(jm, pcc);
		//cout << pcc << endl;
		eo.get_rot_matrix(jm, rm);
		//cout << rm << endl;

		projection(pcc, rm, obj_c, pho_c);
		if (dbg) cout << "pho_c = " << pho_c << endl;

		// get image coords.
		cam.photo_to_image(pho_c, img_c);
		if (dbg) cout << "img_c = " << img_c << endl;

		// get pixel position, column only
		cam.image_to_sensor(img_c, pix_c);
		if (dbg) cout << "pix_c = " << pix_c << endl;

		// get calibrated pixel position
		cam.sensor_to_image(pix_c, cal_c);
		if (dbg) cout << "cal_c = " << cal_c << endl;

		// on left or right side of the sensor line?
		if (img_c.x < cal_c.x)
			ju = jm;
		else
			jl = jm;
		//cout << "ypix=" << pix_c.y << " jm=" << jm <<
		//	" jl=" << jl << " ju=" << ju << endl;
		nit++;
		}
		// *** END OF ITERATION LOOP ***

	pix_c.y = double(jl);

	return true;
}


// back projection by hint method

bool projection_hint(const CCDCamera& cam, const OrientationData& eo,
const DPoint3& obj_c, DPoint& pix_c,
int& nit, double r0, double eps, int max_iter,
bool dbg)
{
	DPoint2 img_c, cal_c;
	DPoint3 pho_c(0., 0., -cam.ck());
	DPoint3 pcc;
	DRotMatrix3 rm;

	// approx. row number
	if (r0 < 0)
		pix_c.y = (eo.nrows() - 1) >> 1; // start at centre row
	else
		pix_c.y = r0; // use approx.

	bool cv = false;	// convergence flag
	nit = 0; // reset iteration counter

	// iteration loop
 	while (nit++ < max_iter && !cv)
		{
		//if (dbg) cout << endl << "Iteration: " << nit << endl;
		//if (dbg) cout << "pix_c = " << pix_c << endl;
		
		// check range
		if (eo.inside(pix_c.y) != true)
			{
			cout << "ERROR: projection_bisection: Index out of EO data range.\n";
			return false;
			}
			
		// get interpolated position and attitude
		eo.get_per_centre(pix_c.y, pcc);
		//cout << pcc << endl;
		eo.get_rot_matrix(pix_c.y, rm);
		//cout << rm << endl;

		projection(pcc, rm, obj_c, pho_c);
		if (dbg) cout << "pho_c = " << pho_c << endl;

		cam.photo_to_image(pho_c, img_c);
		if (dbg) cout << "img_c = " << img_c << endl;

		cam.image_to_sensor(img_c, pix_c);
		if (dbg) cout << "pix_c = " << pix_c << endl;

		cam.sensor_to_image(pix_c, cal_c);
		if (dbg) cout << "cal_c = " << cal_c << endl;

		// x-shift in image space
		double dx = img_c.x - cal_c.x;
		if (dbg) cout << "dx = " << dx << endl;

		// height above ground point
		double hg = pcc.z - obj_c.z;
		if (dbg) cout << "hg = " << hg << endl;

		// x-shift in object space
		double dX = hg / cam.ck() * dx;
		if (dbg) cout << "dX = " << dX << endl;

		DPoint3 d;
		eo.pos_difference(pix_c.y, d);
		if (dbg) cout << "|d| = " << abs(d) << endl;

		// row-shift
		double dr = dX / abs(d);
		if (dbg) cout << "dr = " << dr << endl;

		// apply row-shift
		pix_c.y += dr;

		// limit row position to image boundaries
		if (pix_c.y < 0.)
			pix_c.y = 0.;
		else if (pix_c.y >=  eo.nrows() - 1)
			pix_c.y = eo.nrows() - 2;

		// convergence check
		if (fabs(dr) < eps) cv = true;
		}
		// *** END OF ITERATION LOOP ***

	nit--;

	if (nit < max_iter && cv) return true;

	return false;
}

} // namespace DEP
