/* $Id: dgap.cpp,v 1.14 2008/01/08 16:49:20 cvsroot Exp $ */

/* DGAP - General Analytical Positioning
 * Copyright (C) 2005 Dirk Stallmann
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include <cstdlib>
#include <fstream>
using namespace std;

#include "libdep.h"
using namespace DEP;

#include "timer.h"
using namespace GenLib2;

// name of this program
#define PROGRAM_NAME "dgap"

// version of this program
#define PROGRAM_VERSION "3.2.3"

// build on system
#if defined(__linux__)
	#define SYSTEM_NAME "Linux"
#elif defined(_WIN32)
	#define SYSTEM_NAME "Win32"
#elif defined(__CYGWIN__)
	#define SYSTEM_NAME "CYGWIN"
#else
	#define SYSTEM_NAME "unknown"
#endif

// compiler used
#if defined(__GNUG__)
	#define COMPILER_NAME "GNU C++"
	#define COMPILER_VERSION __VERSION__
#elif defined(_MSC_VER)
	#define COMPILER_NAME "MS-VS"
	#if _MSC_VER == 1200
		#define COMPILER_VERSION "C++ 6.0"
	#elif _MSC_VER == 1300
		#define COMPILER_VERSION "C++ .NET"
	#else
		#define COMPILER_VERSION "unknown"
	#endif
#else
	#define COMPILER_NAME "unknown"
	#define COMPILER_VERSION "unknown"
#endif

const char* const program_name = PROGRAM_NAME;
const char* const program_version = PROGRAM_VERSION;
const char* const system_name = SYSTEM_NAME;
const char* const compiler_name = COMPILER_NAME;
const char* const compiler_version = COMPILER_VERSION;


static void usage(int status=0)
{
	cerr << "Usage: " << program_name << " [options] cfg_file [stem]\n"
		 << "Arguments:\n"
		 << "  cfg_file         Configuration file.\n"
		 << "  stem             Stem name for output files.\n"
		 << "Options:\n"
		 << "  -chk_in          Check input only.\n"
		 << "  -corr            Evaluate correlation coeffs. of parameters.\n"
		 << "  -corr_matrix     Evaluate correlation coeffs. of parameters.\n"
		 << "  -covar           Print covariances of parameters.\n"
		 << "  -db_upari        Print upari-table.\n"
		 << "  -help            Show this text.\n"
		 //<< "  -init_only       Perform initilization only.\n"
		 << "  -nointersec      Don't try intersection to calc. object point coords.\n"
		 << "  -ori_from_dlt    Calculate EO from DLT parameters.\n"
		 << "  -pr_ic           Print image coords.\n"
		 << "  -pr_input        Print all input data.\n"
		 << "  -pr_oc           Print object points after initialization.\n"
		 //<< "  -pr_config       Print image configuration.\n"
		 //<< "  -pr_aps_infl     Print influence of APs on image points.\n"
		 << "  -prec            Evaluate precision of parameters.\n"
		 << "  -resec           Try resection to calc. transf. params.\n"
		 << "  -update_cams     Update cameras by AP sets.\n"
		 << "  -verb            Gives more information.\n"
		 << "  -version         Show program version.\n"
		 << "  -wr_aps file     Write adjusted AP sets.\n"
		 << "  -wr_cam file     Write cameras in file.\n"
		 << "  -wr_eos file     Write adjusted EOs.\n"
		 << "  -wr_mis file     Write adjusted IMU misalignment.\n"
		 << "  -wr_ocs file     Write adjusted object coords.\n"
		 << "  -wr_opa file     Write object point analysis.\n"
		 << "  -wr_res file     Write residuals in file.\n";
	exit(status);
}
	/*
		 << "  -wr_ics file     Write adjusted image coords.\n"
		 << "  -wr_pcs file     Write photo coords.\n"
		 << "  -wr_ueo          Write updated EOs in file\n";
		 << "  -wr_seo stem     Write updated scanner EO ($stem_id.eo).\n";
		 << "  -chk_transf    Check object-pixel transformation.\n"
		 << "  -wr_cgic_adj   Write adjusted CGI config. (stem.cgic_adj)\n"
		 << "  -wr_icor_adj   Write adjusted INS correction (stem.icor_adj)\n"
	*/


static void error(const char* s, const char* s2 = "")
{
	cerr << program_name << ": " << s << ' ' << s2 << endl;
	exit(1);
}


static void version(ostream& os)
{
	os << program_name << " version " << program_version
		<< " system " << system_name
		<< " compiler " << compiler_name << " " << compiler_version
		<< " compiled " << __DATE__ << " " << __TIME__ << endl;
	os.flush();
}


// initialize static data members
LSParams AdjObject::lspars;							// definition of least-squares params.
double OrientationData::shift = 0.;		// global offset for orient. data
OrientationData::IpolMethod OrientationData::ipol = LINEAR;	// interpolation method

int main(int argc, char *argv[])
{
	Timer timer1; timer1.start();

	BundleAdjustment bundle;

	char*	cfg = NULL;			// configuration filename
	char*	stem = NULL;		// stem name
	//string	fname;			// temp. filename
	bool	chk_in = false;		// check input only
	//bool	chk_transf = false;	// check object-pixel transformation
	bool	corr = false;		// evaluate correlation coeffs. of parameters
	bool	corr_matrix = false;	// evaluate correlation coeffs. of parameters
	bool	covar = false;		// print covariances of parameters
	bool	db_upari = false;	// debug upari-table
	//bool	init_only = false;	// perform initilization only
	bool	ori_from_dlt = false;		// calculate orientation from DLT parameters
	bool	pr_ic = false;		// print image coords.
	bool	pr_input = false;	// print all input data
	//bool	pr_config = false;	// print image configuration
	bool	pr_oc = false;		// print object points after initialization
	//bool	pr_aps_infl = false;	// print influence APs on image points
	bool	update_cams = false;	// update cameras by AP sets
	bool	prec = false;		// evaluate precision of parameter
	bool	verb = false;		// verbose
	char*	cam_name = NULL;	// filename of cameras
	//char*	ics_name = NULL;	// filename of adjusted image coords.
	char*	ocs_name = NULL;	// filename of adjusted object coords.
	char*	eos_name = NULL;	// filename of adjusted EOs
	char*	aps_name = NULL;	// filename of adjusted AP sets
	char*	mis_name = NULL;	// filename of adjusted IMU misalignment
	char*	opa_name = NULL;	// filename of object point analysis
	char*	res_name = NULL;	// filename of residuals
	//char*	pcs_name = NULL;	// filename of photo coords.
	//char*	seo_stem = NULL;	// stem name of new (adjusted) EO
	//bool	wr_init_direc = false;	// write initial direction
	//bool	wr_res = false;		// write residuals

	//bool wr_cgic_adj = false;		// write adjusted CGI config.


	if (argc == 1) usage(1);

	for (int i = 1, j = 0; i < argc; i++)
		{
		if (**(argv + i) == '-')
			{
			if (strcmp(argv[i], "-help") == 0)
				usage(0);
			else if (strcmp(argv[i], "-chk_in") == 0)
				chk_in = true;
			//else if (strcmp(argv[i], "-chk_transf") == 0)
			//	chk_transf = true;
			else if (strcmp(argv[i], "-corr") == 0)
				corr = true;
			else if (strcmp(argv[i], "-corr_matrix") == 0)
				corr_matrix = true;
			else if (strcmp(argv[i], "-covar") == 0)
				covar = true;
			else if (strcmp(argv[i], "-db_upari") == 0)
				db_upari = true;
			//else if (strcmp(argv[i], "-init_only") == 0)
			//	init_only = true;
			else if (strcmp(argv[i], "-nointersec") == 0)
				bundle._pars.intersec_flg = false;
			else if (strcmp(argv[i], "-ori_from_dlt") == 0)
				ori_from_dlt = true;
			else if (strcmp(argv[i], "-pr_ic") == 0)
				pr_ic = true;
			else if (strcmp(argv[i], "-pr_input") == 0)
				pr_input = true;
			else if (strcmp(argv[i], "-pr_oc") == 0)
				pr_oc = true;
			//else if (strcmp(argv[i], "-pr_config") == 0)
			//	pr_config = true;
			//else if (strcmp(argv[i], "-pr_aps_infl") == 0)
			//	pr_aps_infl = true;
			else if (strcmp(argv[i], "-prec") == 0)
				prec = true;
			else if (strcmp(argv[i], "-resec") == 0)
				bundle._pars.resec_flg = true;
			else if (strcmp(argv[i], "-update_cams") == 0)
				update_cams = true;
			else if (strcmp(argv[i], "-verb") == 0)
				verb = true;
			else if (strcmp(argv[i], "-version") == 0)
				{
				version(cerr);
				exit(0);
				}
			else if (strcmp(argv[i], "-wr_aps") == 0)
				aps_name = argv[++i];
			else if (strcmp(argv[i], "-wr_cam") == 0)
				cam_name = argv[++i];
			else if (strcmp(argv[i], "-wr_eos") == 0)
				eos_name = argv[++i];
			//else if (strcmp(argv[i], "-wr_ics") == 0)
			//	ics_name = argv[++i];
			else if (strcmp(argv[i], "-wr_mis") == 0)
				mis_name = argv[++i];
			else if (strcmp(argv[i], "-wr_ocs") == 0)
				ocs_name = argv[++i];
			else if (strcmp(argv[i], "-wr_opa") == 0)
				opa_name = argv[++i];
			else if (strcmp(argv[i], "-wr_res") == 0)
				res_name = argv[++i];
//			else if (strcmp(argv[i], "-wr_pcs") == 0)
//				pcs_name = argv[++i];
//			else if (strcmp(argv[i], "-wr_seo") == 0)
//				seo_stem = argv[++i];
//			else if (strcmp(argv[i], "-wr_init_direc") == 0)
//				wr_init_direc = true;
/*
			else if (strcmp(argv[i], "-wr_cgic_adj") == 0)
				wr_cgic_adj = true;
			else if (strcmp(argv[i], "-wr_icor_adj") == 0)
				wr_icor_adj = true;
*/
			else
				error("Illegal option:", argv[i]);
			}
		else
			{
			switch (j++)
				{
				case 0:
					cfg = argv[i];
					break;

				case 1:
					stem = argv[i];
					break;

				default:
					error("Too many arguments given:", argv[i]);
				}
			}
		}

	version(cout);
	cout << endl;

	// print start time
	time_t now = time(0);
	struct tm *tm_now = localtime(&now);
	const size_t sz = 20;
	char str[sz];
	strftime(str, sz, "%Y-%m-%d %H:%M:%S", tm_now);
	cout << "Start: " << str << endl;
	cout << endl;
	
	// read configuration file
	ifstream cfg_in(cfg);
	if (!cfg_in) error("Cannot open configuration file:", cfg);

	if (bundle.read_config(cfg_in) != true)
		error("Reading configuration file failed:", cfg);

	cfg_in.close();

	cout << "Input data" << endl;
	cout << "==========" << endl;
	cout << "Number of ..." << endl;
	cout << "  - images            :   " << bundle._images.size() << endl;
	cout << "  - cameras           :   " << bundle._cameras.size() << endl;
	cout << "  - image points      :   " << bundle.count_image_points() << endl;
	cout << "  - object points     :   " << bundle._obj_points.size() << endl;
	cout << "  - stations          :   " << bundle._stations.size() << endl;
	cout << "  - AP sets           :   " << bundle._ap_sets.size() << endl;
	cout << "  - position corrs.   :   " << bundle._pos_corrs.size() << endl;
	cout << "  - attitude corrs.   :   " << bundle._att_corrs.size() << endl;
	cout << "  - IMU misalignments :   " << bundle._imu_mis.size() << endl;
	cout << "  - orientation data  :   " << bundle._ori_data.size() << endl;
	cout << endl;

	// set option: print image coords.
	bundle._pars.pr_ic = bundle._pars.pr_ic || pr_ic;

	// set option: debug upari-table
	bundle._pars.db_upari = bundle._pars.db_upari || db_upari;

	// print parameters and options
	bundle.print_params(cout);
	bundle.print_ls_params(cout);
	bundle.print_network_params(cout);
	//bundle.print_patb_params(cout);
	//bundle.print_oc_transf(cout);
	cout << endl;

	// print input data
	if (pr_input)
		{
		bundle.print_image_config(cout);
		if (bundle._pars.pr_ic == true) bundle.print_image_coords(cout);
		bundle.print_cameras(cout);
		if (bundle._obj_points.size() > 0) bundle.print_object_points(cout);
		if (bundle._stations.size() > 0) bundle.print_stations(cout);
		if (bundle._ap_sets.size() > 0) bundle.print_ap_sets(cout);
		if (bundle._pos_corrs.size() > 0) bundle.print_pos_corrs(cout);
		if (bundle._att_corrs.size() > 0) bundle.print_att_corrs(cout);
		if (bundle._imu_mis.size() > 0) bundle.print_imu_misalignment(cout);
		if (bundle._ori_data.size() > 0) bundle.print_orientation_data(cout);
		}
	
	if (chk_in) return 0; // escape
	
	// initialize bundle adjustment
	if (bundle.initialize(cout) != true)
		error("Error: Initilize adjustment failed");

	// print image configuration
	//if (pr_config)
	//	bundle.print_image_config(cout);

	// print object points after initialization
	if (pr_oc)
		bundle.print_object_points(cout);
	
	//if (init_only) return 0; // escape
	
	// perform adjustment
	if (bundle.adjustment(cout) != true)
		error("Error: Adjustment failed");

	// print residual image coords.
	bundle.print_residuals(cout);

	// write residual image coords.
	if (res_name != NULL)
		{
		ofstream os(res_name);
		if (!os) error("Cannot open residual output file:", res_name);
		bundle.write_residuals(os);
		cout << "\n*** Residual output file written: " << res_name << endl;
		}

	// print final (adjusted) unknown parameters
	cout << "*** Final (adjusted) unknown parameters ***" << endl;
	cout << endl;
	bundle.print_unknowns(cout);

	// control point residuals
	bundle.ctrl_point_residuals(cout);

	// checkpoint analysis
	bundle.check_point_analysis(cout);

	// write object point analysis
	// a) control point residuals, b) object point differences
	if (opa_name != NULL)
		{
		ofstream os(opa_name);
		if (!os) error("Cannot open object point analysis file:", opa_name);
		bundle.write_object_point_analysis(os);
		cout << "\n*** Object point analysis written: " << opa_name << "\n";
		}

	// print initial direction
	//if (init_direc) bundle.print_init_direction(cout);

	// print influence of APs on image points
	//if (pr_aps_infl) bundle.aps_influence(cout);
	if (bundle._pars.pr_aps_nx > 0 && bundle._pars.pr_aps_ny > 0)
		bundle.aps_influence(cout);
/*
	// write initial direction
	if (wr_init_direc)
		{
		sprintf(fname, "%s.direc", stem);
		ofstream os(fname);
		if (!os) error("Cannot open initial direction file:", fname);
		//bundle.print_init_direction(os);
		cout << "\n*** Initial direction written: " << fname << "\n";
		}
*/

	// inversion
	if (prec || corr || corr_matrix || covar)
		{
		cout << endl;

		// matrix inversion
		//bundle.inversion(&cout);
		// calculate precision
		if (bundle.precision(&cout) != true)
			error("Error: Calculate precision failed");

		// precision for estimated parameters partially with significance test
		if (prec) bundle.precision(cout);

		// print correlations in list form
		if (corr) bundle.correlations_list(cout);

		// print correlations in matrix form
		if (corr_matrix) bundle.correlations_matrix(cout);

		// print covariances
		if (covar) bundle.print_covariances(cout);
		}

	// processing before writing
	// update cameras by AP sets
	if (update_cams)
		{
		bundle.update_cameras();
		cout << "\n*** Cameras updated by AP sets.\n";
		}
/*
	// update scanner EO
	if (seo_stem != NULL)
		{
		cout << "\n*** Update scanner EO ... "; cout.flush();
		//bundle.update_scan_eo(&cout);
		cout << "Done" << endl << endl;
		}
*/

	// IO and EO parameters from DLT parameters
	if (ori_from_dlt)
		bundle.print_orient_from_dlt(cout);
	
	// write final/adjusted data
	// write cameras
	if (cam_name != NULL)
		{
		ofstream os(cam_name);
		if (!os) error("Cannot write camera file:", cam_name);
		bundle.write_cameras(os);
		cout << "\n*** Cameras written: " << cam_name << "\n";
		}

	// write adjusted image coordinates
	//if (ics_name != NULL)
	//	{
	//	ofstream os(ics_name);
	//	if (!os) error("Cannot write adjusted image coords. file:", ics_name);
	//	bundle.write_adjusted image_coords(os);
	//	cout << "\n*** Image coords. written: " << ics_name << "\n";
	//	}

	// write adjusted object coordinates
	if (ocs_name != NULL)
		{
		ofstream os(ocs_name);
		if (!os) error("Cannot write adjusted object coords. file:", ocs_name);
		bundle.write_object_coords(os);
		cout << "\n*** Adjusted object coords. written: " << ocs_name << "\n";
		}

	// write adjusted EOs
	if (eos_name != NULL)
		{
		ofstream os(eos_name);
		if (!os) error("Cannot write adjusted EO file:", eos_name);
		bundle.write_stations(os);
		cout << "\n*** Adjusted station data written: " << eos_name << "\n";
		}

	// write adjusted AP sets
	if (aps_name != NULL)
		{
		ofstream os(aps_name);
		if (!os) error("Cannot write adjusted AP set file:", aps_name);
		bundle.write_ap_sets(os);
		cout << "\n*** Adjusted AP set data written: " << aps_name << "\n";
		}

	// write adjusted IMU misalignment
	if (mis_name)
		{
		ofstream os(mis_name);
		if (!os) error("Cannot write adjusted IMU misalignment file:", mis_name);
		bundle.write_imu_misalignment(os);
		cout << "\n*** Adjusted IMU misalignment written: " << mis_name << "\n";
		}
/*
	// write adjusted CGI configuration
	if (wr_cgic_adj)
		{
		sprintf(fname, "%s.cgic_adj", stem);
		ofstream os(fname);
		if (!os) error("Cannot write adjusted CGI config. file:", fname);
		cgic.write(os, upari);
		cout << "\n*** Adjusted CGI config. written: " << fname << "\n";
		}
*/

	// write photo coordinates
	//if (pcs_name != NULL)
	//	{
	//	ofstream os(pcs_name);
	//	if (!os) error("Cannot write photo coords. file:", pcs_name);
	//	bundle.write_photo_coords(os);
	//	cout << "\n*** Photo coords. written: " << pcs_name << "\n";
	//	}

	// write scanner EO after adjustment
	//if (seo_stem != NULL)
	//	{
		//if (bundle.write_scan_eo(&cout, seo_stem) != OK)
		//	error("Writing scanner EO failed.");
		//bundle.write_scan_eo(seo_stem, &cout);
	//	cout << "\n*** Scanner EO written\n";
	//	}

	// check object-pixel transformation
	//if (chk_transf)
	//	bundle.check_transf(cout);

	cout << " *** Processing completed ***\n\n";
	
	now = time(0);
	tm_now = localtime(&now);
	strftime(str, sz, "%Y-%m-%d %H:%M:%S", tm_now);
	cout << "End: " << str << endl;
	cout << endl;
	double time_all = timer1.stop();
	cout << "Total cpu time (all included): " << time_all << " s\n";

	return 0;
}
